# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for executable code files.
"""


import struct

from . import _common
from . import codeCompression


# Code settings full struct (from nds-bootstrap):
# typedef struct {
#     u32 auto_load_list_offset;
#     u32 auto_load_list_end;
#     u32 auto_load_start;
#     u32 static_bss_start;
#     u32 static_bss_end;
#     u32 compressed_static_end;
#     u32 sdk_version;
#     u32 nitro_code_be;
#     u32 nitro_code_le;
# } module_params_t;


class MainCodeFile:
    """
    Either the main ARM7 code file or the main ARM9 code file.
    """

    class Section:
        """
        A single section within an ARM7 or ARM9 code file. Code not
        technically contained within a section defined in the sections
        table in the code settings block is represented as an "implicit"
        section.
        """

        # There's an implicit first section in the file, which is not
        # defined in the sections table with the others. This attribute
        # will be True if this is that section.
        implicit = False


        def __init__(self, data, ramAddress, bssSize, *, implicit=False):
            self.data = bytearray(data)
            self.ramAddress = ramAddress
            self.bssSize = bssSize
            self.implicit = implicit


        def __str__(self):
            data = _common.shortBytesRepr(self.data)
            imp = ' implicit' if self.implicit else ''
            return f'<code-section at 0x{self.ramAddress:08X}: {data}{imp}>'


        def __repr__(self):
            data = _common.shortBytesRepr(self.data)

            return (f'{type(self).__name__}({data}'
                    f', 0x{self.ramAddress:08X}'
                    f', 0x{self.bssSize:X}'
                    f'{", implicit=True" if self.implicit else ""})')


    sections = None
    ramAddress = 0x00000000
    codeSettingsOffs = 0x00000000


    def __init__(self, data, ramAddress, codeSettingsPointerAddress=None):
        self.sections = []
        self.ramAddress = ramAddress

        data = codeCompression.decompress(data)

        if codeSettingsPointerAddress:
            codeSettingsAddr, = struct.unpack_from(
                '<I', data, codeSettingsPointerAddress - ramAddress - 4)
            self.codeSettingsOffs = codeSettingsAddr - ramAddress
        else:
            # codeSettingsPointerAddress might be None if it's not
            # available, or 0 if the ROM has it set to 0. We need to
            # search for code settings in both of those cases.
            self.codeSettingsOffs = self._searchForCodeSettingsOffs(data)

        if self.codeSettingsOffs is not None:
            copyTableBegin, copyTableEnd, dataBegin = \
                struct.unpack_from('<3I', data, self.codeSettingsOffs)
            copyTableBegin -= ramAddress
            copyTableEnd -= ramAddress
            dataBegin -= ramAddress
        else:
            # The entire file is one implied section
            copyTableBegin = copyTableEnd = 0
            dataBegin = len(data)

        def makeSection(ramAddr, ramLen, fileOffs, bssSize, implicit=False):
            sdata = data[fileOffs : fileOffs + ramLen]
            self.sections.append(self.Section(sdata,
                                              ramAddr,
                                              bssSize,
                                              implicit=implicit))

        # Implicit first section
        makeSection(ramAddress, dataBegin, 0, 0, implicit=True)

        copyTablePos  = copyTableBegin
        while copyTablePos < copyTableEnd:
            secRamAddr, secSize, bssSize = \
                struct.unpack_from('<3I', data, copyTablePos)
            copyTablePos += 12

            makeSection(secRamAddr, secSize, dataBegin, bssSize)

            dataBegin += secSize


    @classmethod
    def fromCompressed(cls, data, *args):
        """
        Create a main code file from compressed code data.

        This function is a bit outdated, since the default constructor
        can now detect code compression. There is no real reason to use
        this function anymore, and it may be removed at some point.

        Parameters are the same as those of the default constructor.
        """
        data = codeCompression.decompress(data)
        return cls(data, *args)


    @classmethod
    def fromSections(cls, sections, ramAddress):
        """
        Create a main code file from a list of sections.
        """
        self = cls(b'', ramAddress)
        self.sections = sections
        return self


    @classmethod
    def fromFile(cls, filePath, ramAddress):
        """
        Load a main code file from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read(), ramAddress)


    def save(self, *, compress=False):
        """
        Generate a bytes object representing this code file.
        """
        data = bytearray()

        for s in self.sections:
            data.extend(s.data)

            # Align to 0x04
            while len(data) % 4:
                data.append(0)

        # These loops are NOT identical!
        # The first one only operates on sections with length != 0,
        # and the second operates on sections with length == 0!

        sectionTable = bytearray()

        for s in self.sections:
            if s.implicit: continue
            if len(s.data) == 0: continue
            sectionTable.extend(
                struct.pack('<3I', s.ramAddress, len(s.data), s.bssSize))

        for s in self.sections:
            if s.implicit: continue
            if len(s.data) != 0: continue
            sectionTable.extend(
                struct.pack('<3I', s.ramAddress, len(s.data), s.bssSize))

        sectionTableOffset = len(data)
        data.extend(sectionTable)

        def setInt(addr, val):
            struct.pack_into('<I', data, addr, val)

        sectionTableAddr = self.ramAddress + sectionTableOffset
        sectionTableEnd = sectionTableAddr + len(sectionTable)

        cso = self.codeSettingsOffs
        if cso is not None:
            setInt(cso + 0x00, sectionTableAddr)
            setInt(cso + 0x04, sectionTableEnd)
            setInt(cso + 0x08, self.ramAddress + len(self.sections[0].data))
        else:
            # Welp, hopefully we only have one section :P
            pass

        if compress:
            data = bytearray(codeCompression.compress(data, True))
            setInt(cso + 0x14, self.ramAddress + len(data))
        else:
            setInt(cso + 0x14, 0)

        return data


    def saveToFile(self, filePath, *, compress=False):
        """
        Generate file data representing this main code file, and save it
        to a filesystem file.
        """
        d = self.save(compress=compress)
        with open(filePath, 'wb') as f:
            f.write(d)


    def _searchForCodeSettingsOffs(self, data):
        """
        Find the offset of the code settings area in the data given.
        Return None if it can't be found.
        """

        # Simple heuristic that works in most arm9.bin's:
        for i in range(0, 0x8000, 4):
            if data[i:i+8] == b'\x21\x06\xC0\xDE\xDE\xC0\x06\x21':
                return i - 0x1C

        # But to support arm7, which lacks that magic, we can fall back
        # to a different heuristic based on the assumption that the code
        # section table will be the very last thing in the code file.
        # Which... isn't a great assumption, but it's the best I've got
        # right now.
        expectedTableEnd = self.ramAddress + len(data)
        expectedTableEndBytes = struct.pack('<I', expectedTableEnd)
        if expectedTableEndBytes in data:
            try:
                match = data.index(expectedTableEndBytes, 0)
            except ValueError:
                match = None
            while match is not None:
                possibleTableStart, = struct.unpack_from('<I', data, match - 4)
                if (possibleTableStart % 4 == 0
                        and (expectedTableEnd - possibleTableStart) % 12 == 0
                        and expectedTableEnd - possibleTableStart < 0x100):
                    # Probably a match
                    return match - 4
                try:
                    match = data.index(expectedTableEndBytes, match + 1)
                except ValueError:
                    match = None


    def __str__(self):
        linesList = [f'<main-code at 0x{self.ramAddress:08X}']

        for s in self.sections:
            linesList.append(f'    {s}')

        linesList.append('>')

        return '\n'.join(linesList)


    def __repr__(self):
        return (f'{type(self).__name__}.fromSections({self.sections!r}'
                f', 0x{self.ramAddress:08X})')


class Overlay:
    """
    An ARM7 or ARM9 code overlay.
    """

    data = None
    ramAddress = 0x00000000
    ramSize = 0
    bssSize = 0
    staticInitStart = 0x00000000
    staticInitEnd = 0x00000000
    fileID = 0
    compressedSize = 0
    flags = 0
    unkAddress = 0x00000000


    def __init__(self, data, ramAddress, ramSize, bssSize, staticInitStart,
            staticInitEnd, fileID, compressedSize, flags):
        self.ramAddress = ramAddress
        self.ramSize = ramSize
        self.bssSize = bssSize
        self.staticInitStart = staticInitStart
        self.staticInitEnd = staticInitEnd
        self.fileID = fileID
        self.compressedSize = compressedSize
        self.flags = flags

        if self.compressed:
            self.data = bytearray(codeCompression.decompress(data))
        else:
            self.data = bytearray(data)


    @property
    def compressed(self):
        return bool(self.flags & 1)
    @compressed.setter
    def compressed(self, value):
        if value:
            self.flags |= 1
        else:
            self.flags &= ~1


    @property
    def verifyHash(self):
        return bool(self.flags & 2)
    @verifyHash.setter
    def verifyHash(self, value):
        if value:
            self.flags |= 2
        else:
            self.flags &= ~2


    def save(self, *, compress=False):
        """
        Generate a bytes object representing this overlay.
        """
        self.ramSize = len(self.data)
        if compress:
            data = codeCompression.compress(self.data, False)
        else:
            data = self.data
        self.compressedSize = len(data)
        self.compressed = compress
        return data


    def __str__(self):
        fields = []
        fields.append('at 0x%08X' % self.ramAddress)
        fields.append(f'file={self.fileID}')
        if self.compressed:
            fields.append('compressed')
        if self.verifyHash:
            fields.append('verify-hash')

        return f'<overlay {" ".join(fields)}>'

    def __repr__(self):
        fields = []
        fields.append(_common.shortBytesRepr(self.data))
        fields.append('0x%08X' % self.ramAddress)
        fields.append('0x%X' % self.ramSize)
        fields.append('0x%X' % self.bssSize)
        fields.append('0x%08X' % self.staticInitStart)
        fields.append('0x%08X' % self.staticInitEnd)
        fields.append(repr(self.fileID))
        fields.append('0x%X' % self.compressedSize)
        fields.append('0x%X' % self.flags)
        return f'{type(self).__name__}({", ".join(fields)})'


def loadOverlayTable(tableData, fileCallback, idsToLoad=None):
    """
    Parse ARM7 or ARM9 overlay table data to create a dictionary of
    Overlays. This is the inverse of saveOverlayTable().
    """
    ovs = {}
    for i in range(0, len(tableData), 32):
        (ovID, ramAddr, ramSize, bssSize, staticInitStart, staticInitEnd,
            fileID, compressedSize_Flags) = struct.unpack_from('<8I', tableData, i)

        if idsToLoad is not None and ovID not in idsToLoad:
            continue

        fileData = fileCallback(ovID, fileID)

        ovs[ovID] = Overlay(fileData, ramAddr, ramSize, bssSize,
            staticInitStart, staticInitEnd, fileID,
            compressedSize_Flags & 0xFFFFFF, compressedSize_Flags >> 24)

    return ovs


def saveOverlayTable(table):
    """
    Generate a bytes object representing this dictionary of Overlays, in
    proper ARM7 or ARM9 overlay table format. This is the inverse of
    loadOverlayTable().
    """
    if not table: return b''

    ovt = bytearray()

    # Ensure that we loop over overlay IDs in order
    for ovId in sorted(table):
        ov = table[ovId]

        values = []
        values.append(ovId)                # 0x00
        values.append(ov.ramAddress)       # 0x04
        values.append(ov.ramSize)          # 0x08
        values.append(ov.bssSize)          # 0x0C
        values.append(ov.staticInitStart)  # 0x10
        values.append(ov.staticInitEnd)    # 0x14
        values.append(ov.fileID)           # 0x18
        values.append(ov.compressedSize | ov.flags << 24) # 0x1C

        ovt.extend(struct.pack('<8I', *values))

    return ovt



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Functions and classes that don't need their own modules.
"""


import enum


class Processor(enum.Enum):
    """
    An enumeration that can be used to distinguish between the Nintendo
    DS's two processors.
    """
    ARM9 = 9
    ARM7 = 7


class WaveType(enum.IntEnum):
    """
    An enumeration that distinguishes between the three types of wave
    data that the Nintendo DS sound hardware understands.
    """
    PCM8 = 0
    PCM16 = 1
    ADPCM = 2


def indexInNamedList(L, name):
    """
    Find the index of the item with a particular name in a list
    containing name-value pairs.
    """
    for i, (nameN, entry) in enumerate(L):
        if nameN == name:
            return i
    raise KeyError(f'{name} not found in the list...')


def findInNamedList(L, name):
    """
    Find the value of the item with a particular name in a list
    containing name-value pairs.
    """
    return L[indexInNamedList(L, name)][1]


def setInNamedList(L, name, value):
    """
    Find the item with a particular name in a list containing name-value
    pairs, and replace its value with a new one. The previous value is
    discarded.
    """
    # Can't assign to tuples, so we need to make a new one
    L[indexInNamedList(L, name)] = (name, value)



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for NARC archives.
"""


import collections
import struct

from . import _common
from . import fnt


def load(data):
    """
    Read NARC data, and create a filename table and a list of files.
    """
    # Read the standard header
    magic, bom, version, filesize, headersize, numblocks = \
        _common.NDS_STD_FILE_HEADER.unpack_from(data, 0)
    if version not in [1, 0x100]:
        raise ValueError(f'Unsupported NARC version: {version}')

    if magic != b'NARC':
        raise ValueError("Wrong magic (should be b'NARC', instead found "
                         f'{magic})')

    # Read the file allocation block
    fatbMagic, fatbSize, fileCount = struct.unpack_from('<4sII', data, 0x10)
    assert fatbMagic == b'FATB'[::-1]

    # Read the file name block
    fntbOffset = 0x10 + fatbSize
    fntbMagic, fntbSize = struct.unpack_from('<4sI', data, fntbOffset)
    assert fntbMagic == b'FNTB'[::-1]

    # Get the data from the file data block before continuing
    fimgOffset = fntbOffset + fntbSize
    fimgMagic, gmifSize = struct.unpack_from('<4sI', data, fimgOffset)
    assert fimgMagic == b'FIMG'[::-1]
    rawDataOffset = fimgOffset + 8

    # Read the file datas
    fileList = []
    for i in range(fileCount):
        startOffset, endOffset = struct.unpack_from('<II', data, 0x1C + 8 * i)
        fileList.append(data[rawDataOffset+startOffset : rawDataOffset+endOffset])

    # Parse the filenames
    names = fnt.load(data[fntbOffset + 8 : fntbOffset + fntbSize])

    return names, fileList


def loadFromFile(filePath):
    """
    Load a NARC archive from a filesystem file, and create a filename
    table and a list of files.
    """
    with open(filePath, 'rb') as f:
        return load(f.read())


def save(filenames, fileList):
    """
    Create a NARC archive from a filename table and a list of files.
    """

    # Prepare the filedata and file allocation table block
    fimgData = bytearray(8)

    fatbData = bytearray()
    fatbData.extend(struct.pack('<4sII',
        b'FATB'[::-1], 0x0C + 8 * len(fileList), len(fileList)))

    # Write data into the FIMG and FAT blocks
    for i, fd in enumerate(fileList):
        startOff = len(fimgData) - 8
        fimgData.extend(fd)
        endOff = startOff + len(fd)
        fatbData.extend(struct.pack('<II', startOff, endOff))
        while len(fimgData) % 4:
            fimgData.append(0)

    # Put the header on the FIMG block
    struct.pack_into('<4sI', fimgData, 0, b'FIMG'[::-1], len(fimgData))

    # Assemble the filename table block
    nameTable = bytearray(fnt.save(filenames))
    while len(nameTable) % 4:
        nameTable.append(0xFF)
    fntbData = struct.pack('<4sI', b'FNTB'[::-1], len(nameTable) + 8) + nameTable

    # Put everything together and return.
    data = bytearray(0x10)
    data.extend(fatbData)
    data.extend(fntbData)
    data.extend(fimgData)
    _common.NDS_STD_FILE_HEADER.pack_into(
        data, 0, b'NARC', 0xFEFF, 1, len(data), 0x10, 3)
    return bytes(data)


def saveToFile(filenames, fileList, filePath):
    """
    Create a NARC archive from a filename table and a list of files, and
    save it to a filesystem file.
    """
    d = save(filenames, fileList)
    with open(filePath, 'wb') as f:
        f.write(d)



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for STRMs.
"""

import struct

from . import WaveType
from . import _common


class STRM:
    """
    A STRM streamed audio file. This is a piece of music, usually used
    for background music or jingles.
    """
    # When saving SDAT, two otherwise identical STRMs will share data
    # only if their dataMergeOptimizationIDs are the same.
    # You can pretty safely ignore this.
    dataMergeOptimizationID = 0

    def __init__(self, file=None, unk02=0, volume=127, priority=64, playerID=0, unk07=0):

        self.unk02 = unk02
        self.volume = volume
        self.priority = priority
        self.playerID = playerID
        self.unk07 = unk07

        if file is not None:
            if not file.startswith(b'STRM'):
                raise ValueError("Wrong magic (should be b'STRM', instead"
                                 f' found {file[:4]})')

            self._initFromData(file)

        else:
            self.waveType = WaveType.PCM8
            self.isLooped = False
            self.channels = []
            self.unk03 = 0
            self.sampleRate = 8000
            self.time = 0
            self.loopOffset = 0
            self.samplesPerBlock = 0
            self.samplesInLastBlock = 0
            self.unk28 = 0
            self.unk2C = 0
            self.unk30 = 0
            self.unk34 = 0
            self.unk38 = 0
            self.unk3C = 0
            self.unk40 = 0
            self.unk44 = 0


    @classmethod
    def fromChannels(cls, channels,
            unk02=0, volume=127, priority=64, playerID=0, unk07=0):
        """
        Create a STRM from a list of channels.
        """
        obj = cls(unk02=unk02, volume=volume, priority=priority,
                  playerID=playerID, unk07=unk07)
        obj.channels = channels
        return obj


    @classmethod
    def fromFile(cls, filePath, *args, **kwargs):
        """
        Load a STRM from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read(), *args, **kwargs)


    def _initFromData(self, file):
        """
        Initialize the STRM from file data.
        """

        magic, bom, version, filesize, headersize, numblocks = \
            _common.NDS_STD_FILE_HEADER.unpack_from(file, 0)
        if version != 0x100:
            raise ValueError(f'Unsupported STRM version: {version}')
        assert magic == b'STRM'

        headMagic, headSize = struct.unpack_from('<4sI', file, 0x10)
        assert headMagic == b'HEAD'

        (waveType, self.isLooped, numChannels, self.unk03,
            self.sampleRate, self.time,
            self.loopOffset,
            numSamples,
            dataOffset,
            numBlocks,
            bytesPerBlock, # per channel
            self.samplesPerBlock, # per channel
            bytesInLastBlock, # per channel
            self.samplesInLastBlock, # per channel
            self.unk28,
            self.unk2C,
            self.unk30,
            self.unk34,
            self.unk38,
            self.unk3C,
            self.unk40,
            self.unk44,
            ) = struct.unpack_from('<B?BB2H16I', file, 0x18)
        assert dataOffset == 0x68
        self.waveType = WaveType(waveType)

        dataOffs = 0x10 + headSize
        dataMagic, dataSize = struct.unpack_from('<4sI', file, dataOffs)
        assert dataMagic == b'DATA'
        data = file[dataOffs + 8 : dataOffs + dataSize]

        isOneBigLongBlock = (numBlocks == 1 and waveType == WaveType.ADPCM)

        self.channels = []
        for _ in range(numChannels):
            self.channels.append([])

        offs = 0
        for bn in range(numBlocks):
            blockSize = bytesPerBlock
            if bn == numBlocks - 1:
                blockSize = bytesInLastBlock

            if isOneBigLongBlock:
                blockSize += 4

            for cn in range(numChannels):
                self.channels[cn].append(data[offs : offs + blockSize])
                offs += blockSize

            while offs % 4: offs += 1


    def save(self, *, updateTime=False):
        """
        Generate file data representing this STRM, and then return that
        data, .unk02, .volume, .priority, .playerID, and .unk07, as a
        6-tuple. This matches the parameters of the default class
        constructor.
        """

        # Figure out the number of blocks, bytes per block, and bytes in
        # the last block (while also checking that these are consistent
        # across channels as they should be)
        numBlocks = bytesPerBlock = bytesInLastBlock = 0

        if self.channels:
            numBlocks = len(self.channels[0])

            # All channels should have the same number of blocks
            for i, blocks in enumerate(self.channels):
                if len(blocks) != numBlocks:
                    raise ValueError(f'Channels 1 and {i + 1} have'
                        f' different numbers of blocks ({numBlocks} vs'
                        f' {len(blocks)})!')

            if numBlocks > 0:
                bytesPerBlock = len(self.channels[0][0])
                bytesInLastBlock = len(self.channels[0][-1])

                # All channels should have all but the final block be
                # bytesPerBlock bytes long, and the last one be
                # bytesInLastBlock long
                for i, blocks in enumerate(self.channels):
                    for j, b in enumerate(blocks[:-1]): # ignore last block
                        if len(b) != bytesPerBlock:
                            raise ValueError('Detected block size is'
                                f' {bytesPerBlock}, but block {j + 1} in'
                                f' channel {i + 1} is {len(b)} bytes'
                                ' long!')

                    b = blocks[-1]
                    if len(b) != bytesInLastBlock:
                        raise ValueError('Detected last block size is'
                            f' {bytesInLastBlock}, but the last block in'
                            f' channel {i + 1} is {len(b)} bytes'
                            ' long!')

        # Construct the wave data
        waveData = bytearray()
        for blocks in zip(*self.channels):
            for b in blocks:
                waveData.extend(b)
            while len(waveData) % 4: waveData.append(0)

        # Construct the file data
        data = bytearray()
        data.extend(_common.NDS_STD_FILE_HEADER.pack(
            b'STRM', 0xFEFF, 0x100, 0x68 + len(waveData), 0x10, 2))

        if updateTime:
            self.time = int(1.0 / self.sampleRate * 16756991 / 32)

        adpcmBlockSizeAdjust = 0
        if self.waveType == WaveType.ADPCM and numBlocks == 1:
            adpcmBlockSizeAdjust = 4

        data.extend(struct.pack(
            '<4sI', b'HEAD', 0x50))
        data.extend(struct.pack('<B?BB2H16I',
            self.waveType, self.isLooped, len(self.channels), self.unk03,
            self.sampleRate, self.time,
            self.loopOffset,
            (numBlocks - 1) * self.samplesPerBlock + self.samplesInLastBlock,
            0x68,
            numBlocks,
            bytesPerBlock - adpcmBlockSizeAdjust,
            self.samplesPerBlock,
            bytesInLastBlock - adpcmBlockSizeAdjust,
            self.samplesInLastBlock,
            self.unk28,
            self.unk2C,
            self.unk30,
            self.unk34,
            self.unk38,
            self.unk3C,
            self.unk40,
            self.unk44,
            ))

        data.extend(struct.pack(
            '<4sI', b'DATA', 8 + len(waveData)))
        data.extend(waveData)

        return (bytes(data),
                self.unk02,
                self.volume,
                self.priority,
                self.playerID,
                self.unk07)


    def saveToFile(self, filePath, *, updateTime=False):
        """
        Generate file data representing this STRM, and save it to a
        filesystem file.
        """
        d = self.save(updateTime=updateTime)[0]
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        if self.channels:
            if all(len(c) == len(self.channels[0]) for c in self.channels):
                num = len(self.channels[0])
                word = 'block' if num == 1 else 'blocks'
                blockInfo = f'of {num} {word} each'
            else:
                blockInfo = 'of differing numbers of blocks each'

            num = len(self.channels)
            word = 'channel' if num == 1 else 'channels'
            channelsInfo = f'[{num} {word} {blockInfo}]'

        else:
            channelsInfo = '[0 channels]'

        try:
            waveType = WaveType(self.waveType).name
        except Exception:
            waveType = f'waveType={hex(self.waveType)}'

        looped = ' looped' if self.isLooped else ''

        return f'<strm {waveType} {self.sampleRate}Hz {channelsInfo} volume={self.volume}{looped}>'


    def __repr__(self):
        c = ['[']
        if self.channels:
            c.append('[')
            if self.channels[0]:
                c.append(_common.shortBytesRepr(self.channels[0][0]))
                if len(self.channels[0]) > 1:
                    c.append(', ...')
            c.append(']')
            if len(self.channels) > 1:
                c.append(', ...')
        c.append(']')
        c = ''.join(c)

        return (f'{type(self).__name__}.fromChannels({c}'
                f', {self.unk02!r}, {self.volume!r}, {self.priority!r}'
                f', {self.playerID!r}, {self.unk07!r})')



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for SSEQs and their sequence events.
"""

import enum
import struct

from . import _common


class SSEQ:
    """
    A SSEQ sound sequence file.
    """
    # When saving SDAT, two otherwise identical SSEQs will share data
    # only if their dataMergeOptimizationIDs are the same.
    # You can pretty safely ignore this.
    dataMergeOptimizationID = 0

    def __init__(self, file=None, unk02=0, bankID=0, volume=127,
                 channelPressure=64, polyphonicPressure=50, playerID=0):
        #        ^
        # Default values here are the most common ones in the NSMB SDAT.

        self.unk02 = unk02
        self.bankID = bankID
        self.volume = volume
        self.channelPressure = channelPressure
        self.polyphonicPressure = polyphonicPressure
        self.playerID = playerID

        if file is not None:
            if not file.startswith(b'SSEQ'):
                raise ValueError("Wrong magic (should be b'SSEQ', instead "
                                 f'found {file[:4]})')

            version, totalFileLen = struct.unpack_from('<HI', file, 6)
            if version != 0x100:
                raise ValueError(f'Unsupported SSEQ version: {version}')

            self._events = []
            dataOffs, = struct.unpack_from('<I', file, 0x18)
            self._eventsData = file[dataOffs:totalFileLen]
            self._parsed = False

        else:

            self._events = []
            self._parsed = True


    @classmethod
    def fromEvents(cls, events, unk02=0, bankID=0, volume=127,
                   channelPressure=64, polyphonicPressure=50, playerID=0):
        """
        Create a new SSEQ object from a list of sequence events.
        """
        obj = cls(None, unk02, bankID, volume, channelPressure,
            polyphonicPressure, playerID)
        obj.events = events
        return obj


    @classmethod
    def fromFile(cls, filePath, *args, **kwargs):
        """
        Load an SSEQ from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read(), *args, **kwargs)


    @property
    def events(self):
        if not self.parsed:
            raise ValueError('You must parse the SSEQ with .parse() before you'
                             ' can access .events!')
        return self._events
    @events.setter
    def events(self, value):
        self._events = value


    @property
    def eventsData(self):
        if self.parsed:
            raise ValueError('You cannot use .eventsData after you have parsed'
                             ' the SSEQ!')
        return self._eventsData
    @eventsData.setter
    def eventsData(self, value):
        self._eventsData = value


    @property
    def parsed(self):
        return self._parsed
    @parsed.setter
    def parsed(self, value):
        raise RuntimeError('SSEQ.parsed is read-only!')


    def parse(self):
        """
        Attempt to process .eventsData to create .events. If successful,
        this switches the SSEQ from the unparsed to the parsed state.
        """
        if self.parsed: return
        self._events, _ = readSequenceEvents(self._eventsData, [])
        self._parsed = True


    def save(self):
        """
        Generate file data representing this SSEQ, and then return
        that data, .unk02, .bankID, .volume, .channelPressure,
        .polyphonicPressure, and .playerID as a 7-tuple. This matches
        the parameters of the default class constructor.
        """
        if self.parsed:
            seqEv, _ = saveSequenceEvents(self.events, [])
        else:
            seqEv = self._eventsData

        file = bytearray(0x1C + len(seqEv))

        _common.NDS_STD_FILE_HEADER.pack_into(file, 0,
            b'SSEQ', 0xFEFF, 0x100, len(file), 0x10, 1)
        struct.pack_into('<4sII', file, 0x10,
            b'DATA', len(file) - 0x10, 0x1C)
        file[0x1C:] = seqEv

        return (file,
                self.unk02,
                self.bankID,
                self.volume,
                self.channelPressure,
                self.polyphonicPressure,
                self.playerID)


    def saveToFile(self, filePath):
        """
        Generate file data representing this SSEQ, and save it to a
        filesystem file.
        """
        d = self.save()[0]
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        fields = (f'bankID={self.bankID} volume={self.volume}'
                  f' channelPressure={self.channelPressure}'
                  f' polyphonicPressure={self.polyphonicPressure}'
                  f' playerID={self.playerID}')
        if not self.parsed:
            return f'<sseq unparsed {fields}>'

        linesList = [f'<sseq {fields}']
        linesList.append(printSequenceEventList(self.events, {}, ' ' * 2))
        linesList.append('>')
        return '\n'.join(linesList)


    def __repr__(self):

        if self.parsed:
            if len(self.events) > 8:
                ev = repr(self.events[:8])[:-1] + ', ...]'
            else:
                ev = repr(self.events)

            return (f'{type(self).__name__}.fromEvents({ev}'
                    f', {self.unk02!r}, {self.bankID!r}'
                    f', {self.volume!r}, {self.channelPressure!r}'
                    f', {self.polyphonicPressure!r}'
                    f', {self.playerID!r})')

        else:
            data = _common.shortBytesRepr(self.eventsData)
            return f'{type(self).__name__}({data})'


class SequenceEvent:
    """
    An abstract base class representing any sequence event in a SSEQ or
    SSAR file.
    """
    dataLength = 1


    def __init__(self, type):
        self.type = type


    def save(self, eventsToOffsets=None):
        """
        Generate data representing this sequence event. This abstract
        base class implementation simply returns a single byte
        containing .type. Subclasses should reimplement this function to
        append their own data to this byte.
        """
        return bytes([self.type])


    @classmethod
    def fromData(cls, type, data, startOffset=0):
        """
        Create an instance of the SequenceEvent subclass this function
        is called on, using a particular type value and reading data
        beginning at some offset. This abstract base class
        implementation simply raises NotImplementedError.
        """
        raise NotImplementedError('SequenceEvent subclasses that can '
            'load themselves from data without context can implement this.')


    def __str__(self):
        return f'<sequence event {hex(self.type)}>'

    def __repr__(self):
        return f'{type(self).__name__}({hex(self.type)})'


def _readVariableLengthInt(data, startOffset, limit=4):
    """
    Read a variable-length integer (as SSEQ encodes them) from `data`
    beginning at `startOffset`, limiting the number of read bytes to
    `limit`.
    
    While the code below looks complicated, the method to read such an
    integer is simple:
    - Read a byte. AND it with 0x7F for the part relevant to the
      integer value.
    - If its MSB (that you just trimmed off) is set, left-shift the int
      value by 7, move on to the next byte and repeat.

    So you read 7 bits as a time for as long as the MSB continues to be
    set.
    """
    offset = startOffset
    value = data[offset] & 0x7F; offset += 1
    length = 0
    while data[offset - 1] & 0x80:
        value <<= 7
        value |= data[offset] & 0x7F; offset += 1; length += 1
        if length > limit:
            raise ValueError('Read variable-length int past its end')
    return value


def _lengthOfVariableLengthInt(x):
    """
    Returns the length of a variable-length integer `x`, as encoded in
    SSEQ. See _readVariableLengthInt() for a description of the format.
    This can be implemented more concisely, but I opted for readability.
    """
    if x < 0:
        raise ValueError(f'Cannot write a negative variable-length int: {x}')
    bits = x.bit_length()
    length = 0
    while bits > 0:
        length += 1
        bits -= 7
    return max(1, length)


def _writeVariableLengthInt(x):
    """
    Find the bytes representing the arbitrarily-large (positive) integer
    `x` in the format used by SSEQ variable-length integer fields.
    See _readVariableLengthInt() for a description of this format.
    """
    if x < 0:
        raise ValueError(f'Cannot write a negative variable-length int: {x}')

    ret = []

    while x:
        value = x & 0x7F
        x >>= 7
        ret.insert(0, value)

    for i, v in enumerate(ret[:-1]):
        ret[i] = v | 0x80

    return bytes(ret)


class NoteSequenceEvent(SequenceEvent):
    """
    A sequence event that plays a note. This class represents sequence
    event types 0x00 through 0x7F; the type value actually determines
    the pitch.
    """

    def __init__(self, type, velocityAndFlag, duration):
        super().__init__(type)
        self.velocity = velocityAndFlag & 0x7F
        self.unknownFlag = bool(velocityAndFlag & 0x80)
        self.duration = duration

    @property
    def name(self):
        return _common.noteName(self.type)

    @property
    def dataLength(self):
        return 2 + _lengthOfVariableLengthInt(self.duration)

    @property
    def pitch(self):
        return self.type
    @pitch.setter
    def pitch(self, value):
        self.type = value

    def save(self, eventsToOffsets=None):
        if self.type < 0:
            raise ValueError(f'Note pitch must be >= 0 (found:'
                             f' {self.type})')
        if self.type > 127:
            raise ValueError(f'Note pitch must be < 128 (found:'
                             f' {self.type})')
        if self.velocity > 127:
            raise ValueError(f'Note velocity must be < 128 (found:'
                             f' {self.velocity})')
        velocityValue = self.velocity | (0x80 if self.unknownFlag else 0)
        return (super().save()
                + bytes([velocityValue])
                + _writeVariableLengthInt(self.duration))

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        velocity = data[startOffset + 1]
        duration = _readVariableLengthInt(data, startOffset + 2)
        return cls(type, velocity, duration)

    def __str__(self):
        flag = ' unknown-flag' if self.unknownFlag else ''
        return f'<{self.name} velocity={self.velocity} duration={self.duration}{flag}>'

    def __repr__(self):
        velocityValue = self.velocity | (0x80 if self.unknownFlag else 0)
        return f'{type(self).__name__}({self.type}, {velocityValue!r}, {self.duration!r})'


class RestSequenceEvent(SequenceEvent):
    """
    A sequence event that causes SSEQ execution to pause for some amount
    of time before moving on. This is sequence event type 0x80.
    """

    def __init__(self, duration):
        super().__init__(0x80)
        self.duration = duration

    @property
    def dataLength(self):
        return 1 + _lengthOfVariableLengthInt(self.duration)

    def save(self, eventsToOffsets=None):
        return super().save() + _writeVariableLengthInt(self.duration)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        duration = _readVariableLengthInt(data, startOffset + 1)
        return cls(duration)

    def __str__(self):
        return f'<rest {self.duration}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.duration!r})'


class InstrumentSwitchSequenceEvent(SequenceEvent):
    """
    A sequence event that causes the track it's placed in to switch to
    using a different instrument (possibly in a different SBNK). This is
    sequence event type 0x81.
    """

    def __init__(self, bankID, instrumentID):
        super().__init__(0x81)
        self.bankID = bankID
        self.instrumentID = instrumentID

    @property
    def dataLength(self):
        return 1 + _lengthOfVariableLengthInt(
            self.bankID << 7 | self.instrumentID)

    def save(self, eventsToOffsets=None):
        value = self.instrumentID & 0x7F | self.bankID << 7
        return super().save() + _writeVariableLengthInt(value)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        value = _readVariableLengthInt(data, startOffset + 1)
        instrumentID = value & 0x7F
        bankID = value >> 7
        return cls(bankID, instrumentID)

    def __str__(self):
        return f'<instrument {self.bankID}/{self.instrumentID}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.bankID!r}, {self.instrumentID!r})'


class BeginTrackSequenceEvent(SequenceEvent):
    """
    A sequence event that declares the location in the sequence event
    data at which a particular track should begin executing. This is
    sequence event type 0x93.
    """
    dataLength = 5

    def __init__(self, trackNumber, firstEvent):
        super().__init__(0x93)
        self.trackNumber = trackNumber
        self.firstEvent = firstEvent

    def save(self, eventsToOffsets=None):
        return (super().save()
                + bytes([self.trackNumber])
                + struct.pack('<I', eventsToOffsets[self.firstEvent])[:3])

    def __str__(self):
        return f'<begin track {self.trackNumber} id={id(self.firstEvent)}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.trackNumber!r}, {self.firstEvent!r} at {id(self.firstEvent)})'


class JumpSequenceEvent(SequenceEvent):
    """
    A sequence event that causes execution of the current track to jump
    to some other location. This is sequence event type 0x94.
    """
    dataLength = 4

    def __init__(self, destination):
        super().__init__(0x94)
        self.destination = destination

    def save(self, eventsToOffsets=None):
        return (super().save()
                + struct.pack('<I', eventsToOffsets[self.destination])[:3])

    def __str__(self):
        return f'<jump id={id(self.destination)}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.destination!r} at {id(self.destination)})'


class CallSequenceEvent(SequenceEvent):
    """
    A sequence event that causes execution of the current track to jump
    to some other location, and pushes the current event's address to a
    return-address stack. This is sequence event type 0x95.
    """
    dataLength = 4

    def __init__(self, destination):
        super().__init__(0x95)
        self.destination = destination

    def save(self, eventsToOffsets=None):
        return (super().save()
                + struct.pack('<I', eventsToOffsets[self.destination])[:3])

    def __str__(self):
        return f'<call id={id(self.destination)}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.destination!r} at {id(self.destination)})'


class RandomSequenceEvent(SequenceEvent):
    """
    A sequence event that causes some other event to execute with a
    randomized argument. This is sequence event type 0xA0.
    """

    def __init__(self, subType, args, randMin, randMax):
        super().__init__(0xA0)
        self.subType = subType
        self.args = args
        self.randMin = randMin
        self.randMax = randMax

    @property
    def dataLength(self):
        return self._dataLength_for_subtype(self.subType)

    @staticmethod
    def _dataLength_for_subtype(subType):
        if subType <= 0x7F:
            return 7

        # This is really really hacky, but... works
        for i in range(5):
            try:
                subEvent = _EVENT_TYPES[subType](0, *([0] * i))
                return 4 + subEvent.dataLength
            except TypeError as e:
                pass
        raise ValueError("Couldn't determine dataLength for"
            ' RandomSequenceEvent for some reason...?'
            f' subType={hex(subType)}')

    def save(self, eventsToOffsets=None):
        return (super().save()
                + bytes([self.subType, *self.args])
                + struct.pack('<hh', self.randMin, self.randMax))

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        subType = data[startOffset + 1]
        length = cls._dataLength_for_subtype(subType)

        args = []
        offset = startOffset + 2
        for i in range(length - 6):
            args.append(data[offset])
            offset += 1

        randMin, randMax = struct.unpack_from('<hh', data, offset)

        return cls(subType, args, randMin, randMax)

    def __str__(self):
        return (f'<random {hex(self.subType)}'
                f' {" ".join([str(x) for x in self.args])}'
                f'{" " if self.args else ""}'
                f'{(self.randMin, self.randMax)}>')

    def __repr__(self):
        return f'{type(self).__name__}({hex(self.subType)}, {self.args!r}, {self.randMin!r}, {self.randMax!r})'


class FromVariableSequenceEvent(SequenceEvent):
    """
    A sequence event that executes some other event with its last
    argument taken from a variable. This is sequence event type 0xA1.
    """

    def __init__(self, subType, variableID, unknown=None):
        super().__init__(0xA1)
        self.subType = subType
        self.variableID = variableID
        self.unknown = unknown

    @property
    def dataLength(self):
        # Original comment on this logic from sseq2mid.c:
        # /* loveemu is a lazy person :P */

        # (loveemu is the author of sseq2mid.)
        # Interpret as you will.

        return 4 if (0xB0 <= self.subType <= 0xBD) else 3

    def save(self, eventsToOffsets=None):
        if self.dataLength == 3:
            return super().save() + struct.pack('<Bh',
                                                self.subType,
                                                self.variableID)
        else:
            if self.unknown is None:
                raise ValueError('FromVariableSequenceEvent: trying to'
                    ' save with unknown, but unknown is None!')
            return super().save() + struct.pack('<Bbbxx',
                                                self.subType,
                                                self.unknown,
                                                self.variableID)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        subType = data[startOffset + 1]
        if 0xB0 <= subType <= 0xBD:
            unknown, variableID = struct.unpack_from('<bb',
                                                     data,
                                                     startOffset + 2)
        else:
            unknown = None
            variableID, = struct.unpack_from('<h', data, startOffset + 2)
        return cls(subType, variableID, unknown)

    def __str__(self):
        unk = f' {self.unknown}' if self.unknown is not None else ''
        return f'<from variable {hex(self.subType)} {self.variableID}{unk}>'

    def __repr__(self):
        unk = f', {self.unknown}' if self.unknown is not None else ''
        return f'{type(self).__name__}({hex(self.subType)}, {self.variableID!r}{unk})'


class IfSequenceEvent(SequenceEvent):
    """
    A sequence event that causes the next event to be skipped if the
    conditional flag is currently false. This is sequence event type
    0xA2.
    """
    def __init__(self):
        super().__init__(0xA2)
    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls()
    def __str__(self):
        return '<if>'
    def __repr__(self):
        return f'{type(self).__name__}()'


def _make_arithmetic_sequence_event_class(typeNum, symbol, name, description):
    """
    Helper function to make a SequenceEvent subclass for the arithmetic
    operations.
    """
    __doc__ = f'A sequence event that {description}. This is sequence event type 0x{typeNum:02X}.'

    def __init__(self, variableID, value):
        SequenceEvent.__init__(self, typeNum)
        self.variableID = variableID
        self.value = value

    def save(self, eventsToOffsets=None):
        return (SequenceEvent.save(self)
                + struct.pack('<Bh', self.variableID, self.value))

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        variableID, value = struct.unpack_from('<Bh', data, startOffset + 1)
        return cls(variableID, value)

    def __str__(self):
        return f'<(var {self.variableID}) {symbol.lower()} {self.value}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.variableID!r}, {self.value!r})'

    return type(name,
                (SequenceEvent,),
                {'__doc__': __doc__,
                 '__init__': __init__,
                 'save': save,
                 'fromData': fromData,
                 '__str__': __str__,
                 '__repr__': __repr__,
                 'dataLength': 4})


VariableAssignmentSequenceEvent = _make_arithmetic_sequence_event_class(0xB0,
    '=', 'VariableAssignmentSequenceEvent', 'sets a variable to a given value')

VariableAdditionSequenceEvent = _make_arithmetic_sequence_event_class(0xB1,
    '+=', 'VariableAdditionSequenceEvent', 'increments a variable by a given value')

VariableSubtractionSequenceEvent = _make_arithmetic_sequence_event_class(0xB2,
    '-=', 'VariableSubtractionSequenceEvent', 'decrements a variable by a given value')

VariableMultiplicationSequenceEvent = _make_arithmetic_sequence_event_class(0xB3,
    '*=', 'VariableMultiplicationSequenceEvent', 'multiplies a variable by a given value')

VariableDivisionSequenceEvent = _make_arithmetic_sequence_event_class(0xB4,
    '/=', 'VariableDivisionSequenceEvent', 'divides a variable by a given value')

VariableShiftSequenceEvent = _make_arithmetic_sequence_event_class(0xB5,
    '[shift]', 'VariableShiftSequenceEvent',
    'performs a "shift" operation on a variable. It\'s unclear what exactly that means')

VariableRandSequenceEvent = _make_arithmetic_sequence_event_class(0xB6,
    '[rand]', 'VariableRandSequenceEvent',
    'performs a "rand" operation on a variable. It\'s unclear what exactly that means')

VariableUnknownB7SequenceEvent = _make_arithmetic_sequence_event_class(0xB7,
    '[unknown op 0xB7]', 'VariableUnknownB7SequenceEvent',
    'performs some currently unknown operation on a variable')

VariableEqualSequenceEvent = _make_arithmetic_sequence_event_class(0xB8,
    '==', 'VariableEqualSequenceEvent',
    'sets the conditional flag to true if the specified variable contains a given value,'
        ' or to false otherwise')

VariableGreaterThanOrEqualSequenceEvent = _make_arithmetic_sequence_event_class(0xB9,
    '>=', 'VariableGreaterThanOrEqualSequenceEvent',
    'sets the conditional flag to true if the specified variable contains a value greater'
        ' than or equal to a given value, or to false otherwise')

VariableGreaterThanSequenceEvent = _make_arithmetic_sequence_event_class(0xBA,
    '>', 'VariableGreaterThanSequenceEvent',
    'sets the conditional flag to true if the specified variable contains a value greater'
        ' than a given value, or to false otherwise')

VariableLessThanOrEqualSequenceEvent = _make_arithmetic_sequence_event_class(0xBB,
    '<=', 'VariableLessThanOrEqualSequenceEvent',
    'sets the conditional flag to true if the specified variable contains a value less'
        ' than or equal to a given value, or to false otherwise')

VariableLessThanSequenceEvent = _make_arithmetic_sequence_event_class(0xBC,
    '<', 'VariableLessThanSequenceEvent',
    'sets the conditional flag to true if the specified variable contains a value less'
        ' than a given value, or to false otherwise')

VariableNotEqualSequenceEvent = _make_arithmetic_sequence_event_class(0xBD,
    '!=', 'VariableNotEqualSequenceEvent',
    'sets the conditional flag to true if the specified variable does not contain'
        ' a given value, or to false otherwise')


def _make_simple_sequence_event_class(typeNum, shortName, name, description):
    """
    Helper function to make a simple SequenceEvent subclass with one
    parameter.
    """
    __doc__ = f'A sequence event {description}. This is sequence event type 0x{typeNum:02X}.'

    def __init__(self, value):
        SequenceEvent.__init__(self, typeNum)
        self.value = value

    def save(self, eventsToOffsets=None):
        return SequenceEvent.save(self) + bytes([self.value])

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(data[startOffset + 1])

    def __str__(self):
        return f'<{shortName.lower()} {self.value}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'

    return type(name,
                (SequenceEvent,),
                {'__doc__': __doc__,
                 '__init__': __init__,
                 'save': save,
                 'fromData': fromData,
                 '__str__': __str__,
                 '__repr__': __repr__,
                 'dataLength': 2})


PanSequenceEvent = _make_simple_sequence_event_class(0xC0,
    'Pan', 'PanSequenceEvent',
    'that sets the stereo panning value for the current track')

TrackVolumeSequenceEvent = _make_simple_sequence_event_class(0xC1,
    'Track volume', 'TrackVolumeSequenceEvent',
    'that sets the volume of the current track')

GlobalVolumeSequenceEvent = _make_simple_sequence_event_class(0xC2,
    'Global volume', 'GlobalVolumeSequenceEvent',
    'that sets the global volume, for all tracks')

TransposeSequenceEvent = _make_simple_sequence_event_class(0xC3,
    'Transpose', 'TransposeSequenceEvent',
    'that causes NoteSequenceEvents following it in the current track to be transposed')

PortamentoSequenceEvent = _make_simple_sequence_event_class(0xC4,
    'Portamento', 'PortamentoSequenceEvent', 'related to portamentos')

PortamentoRangeSequenceEvent = _make_simple_sequence_event_class(0xC5,
    'Portamento range', 'PortamentoRangeSequenceEvent', 'related to portamentos')

TrackPrioritySequenceEvent = _make_simple_sequence_event_class(0xC6,
    'Track priority', 'TrackPrioritySequenceEvent',
    'that sets the priority of the current track')


class MonoPolySequenceEvent(SequenceEvent):
    """
    A sequence event that switches the current track to mono mode or
    poly mode. This is sequence event type 0xC7.
    """
    dataLength = 2

    class Value(enum.IntEnum):
        poly = 0
        mono = 1

    def __init__(self, value):
        super().__init__(0xC7)
        self.value = self.Value(value)

    def save(self, eventsToOffsets=None):
        return super().save() + bytes([self.value])

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(data[startOffset + 1])

    def __str__(self):
        return '<mono>' if self.value else '<poly>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'

# Change the internal name of the MonoPolySequenceEvent.Value enum to
# "MonoPolySequenceEvent.Value"
MonoPolySequenceEvent.Value.__name__ = \
    f'{MonoPolySequenceEvent.__name__}.{MonoPolySequenceEvent.Value.__name__}'


class TieSequenceEvent(SequenceEvent):
    """
    A sequence event that enables or disables "tie" mode on the current
    track. This is sequence event type 0xC8.
    """
    dataLength = 2

    def __init__(self, value):
        super().__init__(0xC8)
        if value in [0, 1]:
            value = bool(value)
        self.value = value

    def save(self, eventsToOffsets=None):
        return super().save() + bytes([self.value])

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(data[startOffset + 1])

    def __str__(self):
        if self.value in [0, 1, False, True]:
            v = 'on' if self.value else 'off'
        else:
            v = self.value
        return f'<tie {v}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'


class PortamentoFromSequenceEvent(SequenceEvent):
    """
    A sequence event related to portamentos. This is sequence event type
    0xC9.
    """
    dataLength = 2

    def __init__(self, value):
        super().__init__(0xC9)
        self.value = value

    def save(self, eventsToOffsets=None):
        return super().save() + bytes([self.value])

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(data[startOffset + 1])

    def __str__(self):
        return f'<portamento from {_common.noteName(self.value)}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'


VibratoDepthSequenceEvent = _make_simple_sequence_event_class(0xCA,
    'Vibrato depth', 'VibratoDepthSequenceEvent',
    'related to vibratos')

VibratoSpeedSequenceEvent = _make_simple_sequence_event_class(0xCB,
    'Vibrato speed', 'VibratoSpeedSequenceEvent',
    'related to vibratos')


class VibratoTypeSequenceEvent(SequenceEvent):
    """
    A sequence event that sets the current vibrato type. This is
    sequence event type 0xCC.
    """
    dataLength = 2

    class Value(enum.IntEnum):
        pitch = 0
        volume = 1
        pan = 2

    def __init__(self, value):
        super().__init__(0xCC)
        self.value = self.Value(value)

    def save(self, eventsToOffsets=None):
        return super().save() + bytes([self.value])

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(data[startOffset + 1])

    def _valueName(self):
        if self.value in [0, 1, 2]:
            return ['pitch', 'volume', 'pan'][self.value]
        else:
            return str(self.value)

    def __str__(self):
        return f'<vibrato type {self._valueName()}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'

# Change the internal name of the VibratoTypeSequenceEvent.Value enum to
# "VibratoTypeSequenceEvent.Value"
VibratoTypeSequenceEvent.Value.__name__ = \
    f'{VibratoTypeSequenceEvent.__name__}.{VibratoTypeSequenceEvent.Value.__name__}'


VibratoRangeSequenceEvent = _make_simple_sequence_event_class(0xCD,
    'Vibrato range', 'VibratoRangeSequenceEvent',
    'related to vibratos')


class PortamentoOnOffSequenceEvent(SequenceEvent):
    """
    A sequence event that enables or disables portamento mode. This is
    sequence event type 0xCE.
    """
    dataLength = 2

    def __init__(self, value):
        super().__init__(0xCE)
        if value in [0, 1]:
            value = bool(value)
        self.value = self.Value(value)

    def save(self, eventsToOffsets=None):
        return super().save() + bytes([self.value])

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(data[startOffset + 1])

    def __str__(self):
        if self.value in [0, 1, False, True]:
            v = 'on' if self.value else 'off'
        else:
            v = self.value
        return f'<portamento {v}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'


PortamentoDurationSequenceEvent = _make_simple_sequence_event_class(0xCF,
    'Portamento duration', 'PortamentoDurationSequenceEvent',
    'related to portamentos')

AttackRateSequenceEvent = _make_simple_sequence_event_class(0xD0,
    'Attack rate', 'AttackRateSequenceEvent',
    'that sets the attack rate for notes in the current track')

DecayRateSequenceEvent = _make_simple_sequence_event_class(0xD1,
    'Decay rate', 'DecayRateSequenceEvent',
    'that sets the decay rate for notes in the current track')

SustainRateSequenceEvent = _make_simple_sequence_event_class(0xD2,
    'Sustain rate', 'SustainRateSequenceEvent',
    'that sets the sustain rate for notes in the current track')

ReleaseRateSequenceEvent = _make_simple_sequence_event_class(0xD3,
    'Release rate', 'ReleaseRateSequenceEvent',
    'that sets the release rate for notes in the current track')


class BeginLoopSequenceEvent(SequenceEvent):
    """
    A sequence event that begins a loop in the current track. This is
    sequence event type 0xD4.

    The end of the loop must be marked by an EndLoopSequenceEvent.
    """
    dataLength = 2

    def __init__(self, loopCount):
        super().__init__(0xD4)
        self.loopCount = loopCount

    def save(self, eventsToOffsets=None):
        return super().save() + bytes([self.loopCount])

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(data[startOffset + 1])

    def __str__(self):
        return f'<begin loop {self.loopCount}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.loopCount!r})'


ExpressionSequenceEvent = _make_simple_sequence_event_class(0xD5,
    'Expression', 'ExpressionSequenceEvent', 'that is unknown')

PrintVariableSequenceEvent = _make_simple_sequence_event_class(0xD6,
    'Print variable', 'PrintVariableSequenceEvent', 'that is unknown')

VibratoDelaySequenceEvent = _make_simple_sequence_event_class(0xE0,
    'Vibrato delay', 'VibratoDelaySequenceEvent',
    'related to vibratos')


class TempoSequenceEvent(SequenceEvent):
    """
    A sequence event that sets the tempo for all tracks in the sequence.
    This is sequence event type 0xE1.
    """
    dataLength = 3

    def __init__(self, value):
        super().__init__(0xE1)
        self.value = value

    def save(self, eventsToOffsets=None):
        return super().save() + struct.pack('<H', self.value)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(struct.unpack_from('<H', data, startOffset + 1)[0])

    def __str__(self):
        return f'<tempo {self.value}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'


class SweepPitchSequenceEvent(SequenceEvent):
    """
    An unknown sequence event type. This is sequence event type 0xE3.
    """
    dataLength = 3

    def __init__(self, value):
        super().__init__(0xE3)
        self.value = value

    def save(self, eventsToOffsets=None):
        return super().save() + struct.pack('<H', self.value)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls(struct.unpack_from('<H', data, startOffset + 1)[0])

    def __str__(self):
        return f'<sweep pitch {self.value}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.value!r})'


class EndLoopSequenceEvent(SequenceEvent):
    """
    A sequence event that ends a loop previously begun with a
    BeginLoopSequenceEvent. This is sequence event type 0xFC.
    """
    def __init__(self):
        super().__init__(0xFC)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls()

    def __str__(self):
        return '<end loop>'

    def __repr__(self):
        return f'{type(self).__name__}()'


class ReturnSequenceEvent(SequenceEvent):
    """
    A sequence event that causes execution of the current track to jump
    back to the most recently encountered CallSequenceEvent. This is
    sequence event type 0xFD.
    """
    def __init__(self):
        super().__init__(0xFD)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls()

    def __str__(self):
        return '<return>'

    def __repr__(self):
        return f'{type(self).__name__}()'


class DefineTracksSequenceEvent(SequenceEvent):
    """
    A sequence event that defines the tracks that will be used in the
    sequence. This is sequence event type 0xFE.
    """
    dataLength = 3

    def __init__(self, trackNumbers):
        super().__init__(0xFE)
        self.trackNumbers = trackNumbers

    def save(self, eventsToOffsets=None):
        tracksBitfield = 0
        for i in range(16):
            if i in self.trackNumbers:
                tracksBitfield |= 1 << i
        return super().save() + struct.pack('<H', tracksBitfield)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        tracksBitfield, = struct.unpack_from('<H', data, startOffset + 1)
        trackNumbers = set()
        for i in range(16):
            if tracksBitfield & (1 << i):
                trackNumbers.add(i)
        return cls(trackNumbers)

    def __str__(self):
        return f'<define tracks {" ".join(str(x) for x in sorted(self.trackNumbers))}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.trackNumbers!r})'


class EndTrackSequenceEvent(SequenceEvent):
    """
    A sequence event that ends execution of the current track. This is
    sequence event type 0xFF.
    """
    def __init__(self):
        super().__init__(0xFF)

    @classmethod
    def fromData(cls, type, data, startOffset=0):
        return cls()

    def __str__(self):
        return '<end track>'

    def __repr__(self):
        return f'{type(self).__name__}()'


class RawDataSequenceEvent(SequenceEvent):
    """
    A dummy sequence event that represents raw binary data that seems to
    be unreachable as far as ndspy can tell.
    """
    @property
    def dataLength(self):
        return len(self.data)

    def __init__(self, data):
        super().__init__(None)
        self.data = data

    def save(self, eventsToOffsets=None):
        return self.data

    def __str__(self):
        return f'<raw data {bytes(self.data)}>'

    def __repr__(self):
        return f'{type(self).__name__}({self.data!r})'


_EVENT_TYPES = {
    0x80: RestSequenceEvent,
    0x81: InstrumentSwitchSequenceEvent,
    0x93: BeginTrackSequenceEvent,
    0x94: JumpSequenceEvent,
    0x95: CallSequenceEvent,
    0xA0: RandomSequenceEvent,
    0xA1: FromVariableSequenceEvent,
    0xA2: IfSequenceEvent,
    0xB0: VariableAssignmentSequenceEvent,
    0xB1: VariableAdditionSequenceEvent,
    0xB2: VariableSubtractionSequenceEvent,
    0xB3: VariableMultiplicationSequenceEvent,
    0xB4: VariableDivisionSequenceEvent,
    0xB5: VariableShiftSequenceEvent,
    0xB6: VariableRandSequenceEvent,
    0xB7: VariableUnknownB7SequenceEvent,
    0xB8: VariableEqualSequenceEvent,
    0xB9: VariableGreaterThanOrEqualSequenceEvent,
    0xBA: VariableGreaterThanSequenceEvent,
    0xBB: VariableLessThanOrEqualSequenceEvent,
    0xBC: VariableLessThanSequenceEvent,
    0xBD: VariableNotEqualSequenceEvent,
    0xC0: PanSequenceEvent,
    0xC1: TrackVolumeSequenceEvent,
    0xC2: GlobalVolumeSequenceEvent,
    0xC3: TransposeSequenceEvent,
    0xC4: PortamentoSequenceEvent,
    0xC5: PortamentoRangeSequenceEvent,
    0xC6: TrackPrioritySequenceEvent,
    0xC7: MonoPolySequenceEvent,
    0xC8: TieSequenceEvent,
    0xC9: PortamentoFromSequenceEvent,
    0xCA: VibratoDepthSequenceEvent,
    0xCB: VibratoSpeedSequenceEvent,
    0xCC: VibratoTypeSequenceEvent,
    0xCD: VibratoRangeSequenceEvent,
    0xCE: PortamentoOnOffSequenceEvent,
    0xCF: PortamentoDurationSequenceEvent,
    0xD0: AttackRateSequenceEvent,
    0xD1: DecayRateSequenceEvent,
    0xD2: SustainRateSequenceEvent,
    0xD3: ReleaseRateSequenceEvent,
    0xD4: BeginLoopSequenceEvent,
    0xD5: ExpressionSequenceEvent,
    0xD6: PrintVariableSequenceEvent,
    0xE0: VibratoDelaySequenceEvent,
    0xE1: TempoSequenceEvent,
    0xE3: SweepPitchSequenceEvent,
    0xFC: EndLoopSequenceEvent,
    0xFD: ReturnSequenceEvent,
    0xFE: DefineTracksSequenceEvent,
    0xFF: EndTrackSequenceEvent,
}


def readSequenceEvents(data, notableOffsets=None):
    """
    Convert raw sequence event data (as seen in SSEQ and SSAR files) to
    a list of SequenceEvent objects. This is the inverse of
    saveSequenceEvents().

    A second list will also be returned that contains the elements from
    the first list that appeared in the input data at the offsets given
    in notableOffsets.
    """
    if notableOffsets is None: notableOffsets = []

    events = {}

    FATE_INPROGRESS = 0
    FATE_RETURN = 1
    FATE_LOOP = 2
    FATE_EOT = 3
    fates = {}

    def parse_at(off):
        offsetsOfMySequentialEvents = []

        while off < len(data):
            if off in fates:
                fate = fates[off]
                if fate == FATE_INPROGRESS:
                    fate = FATE_LOOP
                for off_ in offsetsOfMySequentialEvents:
                    fates[off_] = fate
                return fate

            try:

                type = data[off]

                if type == 0x93: # BeginTrack
                    trackNumber = data[off + 1]
                    firstEventOff, = struct.unpack_from('<I', data, off + 1)
                    firstEventOff >>= 8

                    event = BeginTrackSequenceEvent(trackNumber, None)
                    events[off] = event
                    fates[off] = FATE_INPROGRESS
                    parse_at(firstEventOff)
                    event.firstEvent = events[firstEventOff]

                elif type == 0x94: # Jump
                    destination, = struct.unpack_from('<I', data, off)
                    destination >>= 8

                    event = JumpSequenceEvent(None)
                    events[off] = event
                    fates[off] = FATE_INPROGRESS
                    fate = parse_at(destination)
                    event.destination = events[destination]

                    for off_ in offsetsOfMySequentialEvents:
                        fates[off_] = fate

                    # Should we keep parsing past here? Only if this
                    # is part of an if statement (and thus might be
                    # skipped).
                    x = off - 1
                    while x not in events and x >= 0:
                        x -= 1
                    if x == -1:
                        partOfIfStatement = False
                    else:
                        partOfIfStatement = (events[x].type == 0xA2)

                    if not partOfIfStatement:
                        return fate

                elif type == 0x95: # Call
                    destination, = struct.unpack_from('<I', data, off)
                    destination >>= 8

                    event = CallSequenceEvent(None)
                    events[off] = event
                    fates[off] = FATE_INPROGRESS
                    fate = parse_at(destination)
                    event.destination = events[destination]

                    if fate == FATE_EOT:
                        fates[off] = fate
                        for off_ in offsetsOfMySequentialEvents:
                            fates[off_] = fate
                        return fate
                    elif fate == FATE_RETURN:
                        pass
                    elif fate == FATE_LOOP:
                        fates[off] = fate
                        for off_ in offsetsOfMySequentialEvents:
                            fates[off_] = fate
                        return fate

                elif type == 0xFD: # Return
                    events[off] = ReturnSequenceEvent()
                    fates[off] = FATE_RETURN
                    for off_ in offsetsOfMySequentialEvents:
                        fates[off_] = FATE_RETURN
                    return FATE_RETURN

                elif type == 0xFF: # EoT
                    events[off] = EndTrackSequenceEvent()
                    fates[off] = FATE_EOT
                    for off_ in offsetsOfMySequentialEvents:
                        fates[off_] = FATE_EOT
                    return FATE_EOT

                else:

                    if type <= 0x7F:
                        eventCls = NoteSequenceEvent
                    elif type not in _EVENT_TYPES:
                        raise ValueError(f'Event {hex(type)} unrecognized.')
                    else:
                        eventCls = _EVENT_TYPES[type]

                    event = eventCls.fromData(type, data, off)
                    events[off] = event
                    fates[off] = FATE_INPROGRESS

                offsetsOfMySequentialEvents.append(off)
                off += event.dataLength

            except (struct.error, IndexError):
                raise EOFError('Reached EoF of sequence.')
        raise EOFError('Reached EoF of sequence.')

    starts = notableOffsets
    if not starts: starts = [0]
    for start in starts:
        ultimateFate = parse_at(start)
        assert ultimateFate in (FATE_EOT, FATE_LOOP)

    eventsList = []
    i = 0
    while i < len(data):
        if i in events:
            eventsList.append(events[i])
            i += events[i].dataLength
        else:
            j = i
            while j not in events and j < len(data):
                j += 1
            eventsList.append(RawDataSequenceEvent(data[i:j]))
            i = j

    notableEvents = [events[off] for off in notableOffsets]

    return eventsList, notableEvents


def saveSequenceEvents(events, notableEvents=None):
    """
    Convert a list of SequenceEvent objects to raw sequence event data.
    This is the inverse of readSequenceEvents().

    A second list will also be returned that contains the offsets in the
    output data of the elements from notableEvents.
    """
    if notableEvents is None: notableEvents = []
    
    events2Offsets = {}

    off = 0
    for e in events:
        events2Offsets[e] = off
        off += e.dataLength
    data = bytearray(off)

    for event, offset in events2Offsets.items():
        eData = event.save(events2Offsets)
        data[offset : offset + len(eData)] = eData

    notableOffsets = [events2Offsets[e] for e in notableEvents]

    return data, notableOffsets


def printSequenceEventList(events, labels=None, linePrefix=''):
    """
    Produce a string representation of a list of sequence events. You
    can optionally provide a dictionary of labels to mark certain
    events, and a prefix string that will be prepended to every line.
    """
    if labels is None: labels = {}

    warningsList = []
    linesList = []

    maxNameLen = 0
    i2SfxName = {}
    for name, event in labels.items():
        if event in events:
            i = events.index(event)
        elif event is None:
            linesList.append(f'{linePrefix}{name}: (none)')
            continue
        else:
            warningsList.append(f'{linePrefix}(WARNING: {event!r} ({name}) not in events list!)')
            continue
        if name is None: name = str(i)
        if i not in i2SfxName:
            i2SfxName[i] = name
        else:
            i2SfxName[i] += ', ' + name
        maxNameLen = max(maxNameLen, len(i2SfxName[i]))
    maxNameLen = min(maxNameLen, 48)
    maxNumLen = len(str(len(events) - 1))

    maxNameLen += 1 # for the ":" at the end

    def getDestinationStr(primaryEvent, destination):
        if destination in events:
            return str(events.index(destination))
        else:
            warningsList.append(f'{linePrefix}(WARNING: {destination!r} (target of {primaryEvent!r}) not in events list!)')
            return 'NOWHERE'

    def printEvent(e):
        label = i2SfxName[i] + ':' if i in i2SfxName else ''
        labelLst = []
        while label:
            labelLst.append(label[:maxNameLen])
            label = label[maxNameLen:]
        if labelLst:
            labelLst[-1] = f'{labelLst[-1]:{maxNameLen}}'
        else:
            labelLst = [' ' * maxNameLen]
        label = f'\n{linePrefix}'.join(labelLst)

        if isinstance(e, JumpSequenceEvent):
            etext = f'<jump @{getDestinationStr(e, e.destination)}>'
        elif isinstance(e, CallSequenceEvent):
            etext = f'<call @{getDestinationStr(e, e.destination)}>'
        elif isinstance(e, BeginTrackSequenceEvent):
            etext = f'<begin track {e.trackNumber} @{getDestinationStr(e, e.firstEvent)}>'
        else:
            etext = str(e)

        return f'{linePrefix}{label}  [{i:{maxNumLen}}] {etext},'

    for i, e in enumerate(events):
        linesList.append(printEvent(e))

    # Remove last ","
    if events:
        linesList[-1] = linesList[-1][:-1]

    return '\n'.join(warningsList + linesList)



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.


import struct

import crcmod


# Nintendo DS standard file header:
NDS_STD_FILE_HEADER = struct.Struct('<4sHHIHH')
# - Magic
# - BOM
# - Version (generally 1)
# - File size (including this header)
# - Size of this header (16)
# - Number of blocks


def crc16(data):
    return crcmod.predefined.mkCrcFun('modbus')(data)


def loadNullTerminatedStringFrom(
        data, offset, charWidth=1, encoding='latin-1'):
    """
    Load a null-terminated string from data at offset, with the options
    given
    """
    end = data.find(b'\0' * charWidth, offset)
    return data[offset:end].decode(encoding)


def listFind(L, *args):
    try:
        return L.index(*args)
    except ValueError:
        return -1


def noteName(value):
    note = value - 60
    letter = 'ccddeffggaab'[note % 12]
    sharp  = ' ♯ ♯  ♯ ♯ ♯ '[note % 12].strip()
    if note < -12:
        letter = letter.upper()
    accents = ''
    while note >= 0:
        accents += "'"
        note -= 12
    while note < -24:
        accents += '͵'
        note += 12
    return letter + sharp + accents


def isIterable(object):
    """
    Check whether an object is iterable or not.
    """
    # https://stackoverflow.com/a/1952481
    try:
        iter(object)
        return True
    except TypeError:
        return False


def shortBytesRepr(data):
    """
    Like bytes.__repr__(), but will truncate large amounts of data.
    Will also take advantage of octal encoding to make the output more
    compact.
    """
    MAX_LEN = 0x20

    dataTrunc = data[:MAX_LEN]
    r = ["b'"]
    for b in dataTrunc:
        if b < 8:
            r.append(f'\\{b}')
        else:
            r.append(repr(bytes([b]))[2:-1])
    r.append("'")

    final = ''.join(r)

    if len(data) > MAX_LEN:
        return final + '...'
    else:
        return final


def shortColorsListRepr(colors):
    """
    Like shortBytesRepr(), but for lists of colors.
    """
    MAX_COLORS = 20
    colorsTrunc = colors[:MAX_COLORS]
    hexColors = (f'0x{c:04X}' for c in colorsTrunc)
    ellipsis = ', ...' if len(colors) > MAX_COLORS else ''
    return f'[{", ".join(hexColors)}{ellipsis}]'


def loadInfoBlock(data, off, expectedEntryLengths=None):
    """
    Read an info block.
    These are used extensively in NSBMD and NSBTX files.
    """

    # Read overall header
    count, infoBlockSize = struct.unpack_from('<xBH', data, off)
    off += 4

    # Read unk block header and get array pointer
    unkHeaderLen, unkLen, unkUnk04 = \
        struct.unpack_from('<HHI', data, off)
    off += 8
    assert unkHeaderLen == 8
    assert unkLen == 12 + 4 * count
    assert unkUnk04 == 0x17F
    unkBlockOff = off
    off += 4 * count

    # Read entries block header and get array pointer
    eachEntryLen, entriesLen = \
        struct.unpack_from('<HH', data, off)
    if expectedEntryLengths is not None:
        assert eachEntryLen == expectedEntryLengths
    assert entriesLen == 4 + count * eachEntryLen
    off += 4
    entryOff = off
    off += eachEntryLen * count

    # Get names array pointer
    nameOff = off

    # Gather entries into a list
    entries = []
    for i in range(count):
        unk00, unk02 = struct.unpack_from('<HH', data, unkBlockOff)
        entry = data[entryOff : entryOff+eachEntryLen]
        name = data[nameOff : nameOff+16].rstrip(b'\0').decode('latin-1')

        entries.append((name, unk00, unk02, entry))

        unkBlockOff += 4
        entryOff += eachEntryLen
        nameOff += 16

    return entries


def saveInfoBlock(entries, entryLengths):
    """
    The inverse of loadInfoBlock().
    entryLengths is required, so the code can properly handle the case
    of a completely empty entries list.
    """
    data = bytearray()

    # Create the overall header
    # (overall size value will be inserted at the end)
    data.extend(struct.pack('<xBH', len(entries), 0))

    # Write the unk block
    data.extend(struct.pack('<HHI', 8, 12 + 4 * len(entries), 0x17F))
    for e in entries:
        data.extend(struct.pack('<HH', e[1], e[2]))

    # Write the entries block
    data.extend(struct.pack('<HH', entryLengths, 4 + len(entries) * entryLengths))
    for e in entries:
        if len(e[3]) != entryLengths:
            raise ValueError('One of the entries is not of length'
                f' {entryLengths} like it should be'
                f' (actual length is {len(e[3])})')
        data.extend(e[3])

    # Write the names block
    for e in entries:
        if e[0] is None:
            data.extend(b'\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0')
        else:
            data.extend(e[0].encode('latin-1').ljust(16, b'\0'))

    # Insert the overall size, now that we know what it is
    struct.pack_into('<H', data, 2, len(data))

    return bytes(data)



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for SWAR sound wave archives.
"""

import struct

from . import WaveType
from . import _common
from . import soundWave


class SWAR:
    """
    A SWAR sound wave archive.
    """
    # When saving SDAT, two otherwise identical SWARs will share data
    # only if their dataMergeOptimizationIDs are the same.
    # You can pretty safely ignore this.
    dataMergeOptimizationID = 0

    def __init__(self, file=None, unk02=0):
        self.unk02 = unk02
        self.waves = []

        if file is not None:
            if not file.startswith(b'SWAR'):
                raise ValueError("Wrong magic (should be b'SWAR', instead "
                                 f'found {file[:4]})')

            self._initFromData(file)


    def _initFromData(self, data):
        """
        Initialize the SWAR from file data.
        """
        magic, bom, version, filesize, headersize, numblocks = \
            _common.NDS_STD_FILE_HEADER.unpack_from(data, 0)
        if version != 0x100:
            raise ValueError(f'Unsupported SWAR version: {version}')
        assert magic == b'SWAR'

        dataMagic, dataSize, swavCount = \
            struct.unpack_from('<4sI32xI', data, 0x10)
        assert dataMagic == b'DATA'

        dataArrayPos = 0x3C
        for i in range(swavCount):
            swavOffset, nextOffset = \
                struct.unpack_from('<II', data, dataArrayPos)
            dataArrayPos += 4
            if i == swavCount - 1:
                # DON'T use len(data) here -- there can sometimes be
                # extra pad at the end, which will mess this up!
                nextOffset = filesize

            swavData = data[swavOffset:nextOffset]

            waveType, loopFlag, sampleRate, time, loopOffset, totalLength = \
                struct.unpack_from('<B?3HI', swavData)

            swav = soundWave.SWAV.fromData(swavData[0xC:],
                                           waveType=WaveType(waveType),
                                           isLooped=loopFlag,
                                           sampleRate=sampleRate,
                                           time=time,
                                           loopOffset=loopOffset,
                                           totalLength=loopOffset + totalLength)
            self.waves.append(swav)


    @classmethod
    def fromWaves(cls, waves, unk02=0):
        """
        Create a SWAR from a list of SWAVs.
        """
        obj = cls(unk02=unk02)
        obj.waves = waves
        return obj


    @classmethod
    def fromFile(cls, filePath, unk02=0):
        """
        Load a SWAR from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read(), unk02=0)


    def save(self):
        """
        Generate file data representing this SWAR, and then return that
        data and .unk02 as a pair (2-tuple). This matches the parameters
        of the default class constructor.
        """
        fileLen = 0x3C + sum(4 + 0xC + len(s.data) for s in self.waves)

        data = bytearray()
        data.extend(_common.NDS_STD_FILE_HEADER.pack(
            b'SWAR', 0xFEFF, 0x100, fileLen, 0x10, 1))
        data.extend(struct.pack(
            '<4sI32xI', b'DATA', fileLen - 0x10, len(self.waves)))

        lenOffsetsTable = 4 * len(self.waves)
        offsets = []
        swavsTable = bytearray()

        for swav in self.waves:
            offsets.append(0x3C + lenOffsetsTable + len(swavsTable))
            swavData = swav.save()[0x18:] # Chop off the file header
                                          # but not the SWAVInfo struct
            swavsTable.extend(swavData)

        offsetsTable = struct.pack(f'<{len(offsets)}I', *offsets)
        data.extend(offsetsTable)
        data.extend(swavsTable)

        return (data, self.unk02)


    def saveToFile(self, filePath):
        """
        Generate file data representing this SWAR, and save it to a
        filesystem file.
        """
        d = self.save()[0]
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        linesList = [f'<swar']

        maxNumLen = len(str(len(self.waves) - 1))

        for i, wav in enumerate(self.waves):
            linesList.append(f'    [{i:{maxNumLen}}] {wav}')

        linesList.append('>')
        return '\n'.join(linesList)


    def __repr__(self):
        return f'{type(self).__name__}.fromWaves({self.waves!r})'



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for ROMs.
"""


import math
import struct

from . import _common
from . import code
from . import fnt as fntLib

ICON_BANNER_LEN = 0x840


class NintendoDSRom:
    """
    A Nintendo DS ROM file (.nds).
    """

    def __init__(self, data=None):
        super().__init__()

        if data is None:
            self._initAsNew()
        else:
            self._initFromData(data)


    def _initAsNew(self):
        """
        Initialize this ROM with default values.
        """

        self.name = b''
        self.idCode = b'####'
        self.developerCode = b'\0\0'
        self.unitCode = 0
        self.encryptionSeedSelect = 0
        self.deviceCapacity = 9
        self.pad015 = 0
        self.pad016 = 0
        self.pad017 = 0
        self.pad018 = 0
        self.pad019 = 0
        self.pad01A = 0
        self.pad01B = 0
        self.pad01C = 0
        self.region = 0
        self.version = 0
        self.autostart = 0
        self.arm9EntryAddress = 0x2000800
        self.arm9RamAddress = 0x2000000
        self.arm7EntryAddress = 0x2380000
        self.arm7RamAddress = 0x2380000
        self.normalCardControlRegisterSettings = 0x0416657
        self.secureCardControlRegisterSettings = 0x81808f8
        self.secureAreaChecksum = 0x0000
        self.secureTransferDelay = 0x0D7E
        self.arm9CodeSettingsPointerAddress = 0
        self.arm7CodeSettingsPointerAddress = 0
        self.secureAreaDisable = b'\0' * 8
        self.pad088 = b'\0' * 0x38
        self.nintendoLogo = (b'$\xff\xaeQi\x9a\xa2!=\x84\x82\n\x84\xe4\t\xad'
            b"\x11$\x8b\x98\xc0\x81\x7f!\xa3R\xbe\x19\x93\t\xce \x10FJJ\xf8'1"
            b'\xecX\xc7\xe83\x82\xe3\xce\xbf\x85\xf4\xdf\x94\xceK\t\xc1\x94V'
            b"\x8a\xc0\x13r\xa7\xfc\x9f\x84Ms\xa3\xca\x9aaX\x97\xa3'\xfc\3\x98"
            b'v#\x1d\xc7a\3\4\xaeV\xbf8\x84\0@\xa7\x0e\xfd\xffR\xfe\3o\x950'
            b'\xf1\x97\xfb\xc0\x85`\xd6\x80%\xa9c\xbe\3\1N8\xe2\xf9\xa24\xff'
            b'\xbb>\3Dx\0\x90\xcb\x88\x11:\x94e\xc0|c\x87\xf0<\xaf\xd6%\xe4'
            b'\x8b8\n\xacr!\xd4\xf8\7')
        self.debugRomAddress = 0
        self.pad16C = b'\0' * 0x94
        self.pad200 = b'\0' * 0x3E00

        self.arm9 = b''
        self.arm9PostData = b''
        self.arm7 = b''
        self.arm9OverlayTable = b''
        self.arm7OverlayTable = b''
        self.iconBanner = b''
        self.debugRom = b''

        self.filenames = fntLib.Folder()
        self.files = []
        self.sortedFileIds = []


    def _initFromData(self, data):
        """
        Initialize this ROM from existing data.
        """
        # I could read the header as one huge struct,
        # but... no.

        data = bytearray(data)
        if len(data) < 0x200:
            data.extend(b'\0' * (0x200 - len(data)))
            assert len(data) == 0x200

        headerOffset = 0
        def readRaw(length):
            nonlocal headerOffset
            retVal = data[headerOffset : headerOffset+length]
            headerOffset += length
            return retVal
        def read8():
            nonlocal headerOffset
            retVal = data[headerOffset]
            headerOffset += 1
            return retVal
        def read16():
            nonlocal headerOffset
            retVal, = struct.unpack_from('<H', data, headerOffset)
            headerOffset += 2
            return retVal
        def read32():
            nonlocal headerOffset
            retVal, = struct.unpack_from('<I', data, headerOffset)
            headerOffset += 4
            return retVal

        assert headerOffset == 0
        self.name = readRaw(12).rstrip(b'\0')
        self.idCode = readRaw(4)
        self.developerCode = readRaw(2)
        self.unitCode = read8()
        self.encryptionSeedSelect = read8()
        self.deviceCapacity = read8()
        assert headerOffset == 0x15
        self.pad015 = read8()
        self.pad016 = read8()
        self.pad017 = read8()
        self.pad018 = read8()
        self.pad019 = read8()
        self.pad01A = read8()
        self.pad01B = read8()
        self.pad01C = read8()
        self.region = read8()
        self.version = read8()
        self.autostart = read8()
        assert headerOffset == 0x20
        arm9Offset = read32()
        self.arm9EntryAddress = read32()
        self.arm9RamAddress = read32()
        arm9Len = read32()
        arm7Offset = read32()
        self.arm7EntryAddress = read32()
        self.arm7RamAddress = read32()
        arm7Len = read32()
        assert headerOffset == 0x40
        fntOffset = read32()
        fntLen = read32()
        fatOffset = read32()
        fatLen = read32()
        arm9OvTOffset = read32()
        arm9OvTLen = read32()
        arm7OvTOffset = read32()
        arm7OvTLen = read32()
        assert headerOffset == 0x60
        self.normalCardControlRegisterSettings = read32()
        self.secureCardControlRegisterSettings = read32()
        iconBannerOffset = read32()
        self.secureAreaChecksum = read16() # TODO: Actually recalculate
                                           # this upon saving.
        self.secureTransferDelay = read16()
        assert headerOffset == 0x70
        self.arm9CodeSettingsPointerAddress = read32()
        self.arm7CodeSettingsPointerAddress = read32()
        self.secureAreaDisable = readRaw(8)
        assert headerOffset == 0x80
        romSizeOrRsaSigOffset = read32()
        headerSize = read32()
        self.pad088 = readRaw(0x38)
        self.nintendoLogo = readRaw(0x9C)
        nintendoLogoChecksum = read16()
        headerChecksum = read16()
        assert headerOffset == 0x160
        debugRomOffset = read32()
        debugRomSize = read32()
        self.debugRomAddress = read32()
        self.pad16C = readRaw(0x94)
        assert headerOffset == 0x200
        self.pad200 = data[0x200 : min(arm9Offset, len(data))]

        # Read the RSA signature file
        realSigOffset = 0
        if len(data) >= 0x1004:
            realSigOffset, = struct.unpack_from('<I', data, 0x1000)
        if not realSigOffset and len(data) > romSizeOrRsaSigOffset:
            realSigOffset = romSizeOrRsaSigOffset
        self.rsaSignature = b''
        if realSigOffset:
            self.rsaSignature = data[realSigOffset : min(len(data), realSigOffset + 0x88)]

        # Read arm9, arm7, FNT, FAT, overlay tables, icon banner
        self.arm9 = data[arm9Offset : arm9Offset+arm9Len]
        self.arm7 = data[arm7Offset : arm7Offset+arm7Len]
        fnt = data[fntOffset : fntOffset+fntLen]
        fat = data[fatOffset : fatOffset+fatLen]
        self.arm9OverlayTable = data[
            arm9OvTOffset : arm9OvTOffset + arm9OvTLen]
        self.arm7OverlayTable = data[
            arm7OvTOffset : arm7OvTOffset + arm7OvTLen]
        if iconBannerOffset:
            self.iconBanner = \
                data[iconBannerOffset : iconBannerOffset + ICON_BANNER_LEN]
        else:
            self.iconBanner = b''
        if debugRomOffset:
            self.debugRom = \
                data[debugRomOffset : debugRomOffset + debugRomSize]
        else:
            self.debugRom = b''

        # Read the small amount of data immediately following arm9
        # No idea what this is, though...
        # Probably related to the "code settings" stuff in code.py.
        arm9PostData = bytearray()
        arm9PostDataOffset = arm9Offset+arm9Len
        while (data[arm9PostDataOffset:arm9PostDataOffset+4]
                == b'\x21\x06\xC0\xDE'):
            arm9PostData.extend(data[arm9PostDataOffset:arm9PostDataOffset+12])
            arm9PostDataOffset += 12
        self.arm9PostData = arm9PostData

        # Read the filename table
        if fnt:
            self.filenames = fntLib.load(fnt)
        else:
            self.filenames = fntLib.Folder()

        # Read files
        self.files = []
        self.sortedFileIds = []
        if fat:
            offset2Id = {}
            for i in range(len(fat) // 8):
                startOffset, endOffset = struct.unpack_from('<II', fat, 8 * i)
                self.files.append(data[startOffset:endOffset])
                offset2Id[startOffset] = i
            for off in sorted(offset2Id):
                self.sortedFileIds.append(offset2Id[off])


    @classmethod
    def fromFile(cls, filePath):
        """
        Load a ROM from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read())


    def save(self, *, updateDeviceCapacity=False):
        """
        Generate file data representing this ROM.
        """

        fileOffsets = {}

        # The header will be filled in at the end.
        data = bytearray(0x200)

        def align(alignment, fill=b'\0'):
            if len(data) % alignment:
                extra = len(data) % alignment
                needed = alignment - extra
                data.extend(fill * needed)

        # Add post-header padding
        data.extend(self.pad200)
        align(0x4000)

        # Pack arm9
        arm9Offset = len(data)
        data.extend(self.arm9)
        data.extend(self.arm9PostData)
        align(0x200, b'\xFF')

        # Pack arm9 overlay table
        if self.arm9OverlayTable:
            arm9OvTOffset = len(data)
            data.extend(self.arm9OverlayTable)
            align(0x200, b'\xFF')
        else:
            arm9OvTOffset = 0

        # Pack arm9 overlays
        for i in range(0, len(self.arm9OverlayTable), 32):
            fileId, = struct.unpack_from('<I', self.arm9OverlayTable, i + 0x18)
            fileOffsets[fileId] = len(data)
            data.extend(self.files[fileId])
            align(0x200, b'\xFF')

        # Pack arm7
        arm7Offset = len(data)
        data.extend(self.arm7)
        align(0x200, b'\xFF')

        # Pack arm7 overlay table
        if self.arm7OverlayTable:
            arm7OvTOffset = len(data)
            data.extend(self.arm7OverlayTable)
            align(0x200, b'\xFF')
        else:
            arm7OvTOffset = 0

        # Pack arm7 overlays
        for i in range(0, len(self.arm7OverlayTable), 32):
            fileId, = struct.unpack_from('<I', self.arm7OverlayTable, i + 0x18)
            fileOffsets[fileId] = len(data)
            data.extend(self.files[fileId])
            align(0x200, b'\xFF')

        # Pack the filename table
        fntOffset = len(data)
        fnt = fntLib.save(self.filenames)
        data.extend(fnt)
        align(0x200, b'\xFF')

        # Leave some empty space for the file allocation table -- we'll
        # fill in the real values later
        fatOffset = len(data)
        data.extend(b'\0' * 8 * len(self.files))
        align(0x200, b'\xFF')

        # Pack the icon/banner
        if self.iconBanner:
            assert len(self.iconBanner) == ICON_BANNER_LEN
            iconBannerOffset = len(data)
            data.extend(self.iconBanner)
            align(0x200, b'\xFF')
        else:
            iconBannerOffset = 0

        # Pack the debug rom
        # I don't know if this is really where it would go, but it seems
        # to be the logical place to put it...
        if self.debugRom:
            debugRomOffset = len(data)
            data.extend(self.debugRom)
            align(0x200, b'\xFF')
        else:
            debugRomOffset = 0

        # Pack the rest of the files
        def iterFilenums():
            for fileNum in self.sortedFileIds:
                if fileNum not in fileOffsets and fileNum < len(self.files):
                    yield fileNum
            for fileNum in range(len(self.files)):
                if fileNum not in fileOffsets:
                    yield fileNum

        for fileNum in iterFilenums():
            # Align before instead of after, so that there's no extra
            # padding after the last file
            align(0x200, b'\xFF')
            fileOffsets[fileNum] = len(data)
            data.extend(self.files[fileNum])

        # Pack the FAT
        for i, file in enumerate(self.files):
            assert i in fileOffsets
            startOffset = fileOffsets[i]
            endOffset = startOffset + len(file)
            struct.pack_into('<II', data, fatOffset + 8 * i, startOffset, endOffset)

        # Pack the RSA signature
        align(0x20)
        rsaSignatureOffset = len(data)
        data.extend(self.rsaSignature)

        # We need to do this for compatibility with NSMBe
        struct.pack_into('<I', data, 0x1000, rsaSignatureOffset)

        # Now that we know how large the ROM data is, we can update the
        # device capacity value
        if updateDeviceCapacity:
            self.deviceCapacity = math.ceil(math.log2(len(data))) - 17

        # Now that all the offsets and stuff are determined, write the
        # header data
        headerOffset = 0
        def writeRaw(value):
            nonlocal headerOffset
            data[headerOffset : headerOffset+len(value)] = value
            headerOffset += len(value)
        def write8(value):
            nonlocal headerOffset
            data[headerOffset] = value
            headerOffset += 1
        def write16(value):
            nonlocal headerOffset
            struct.pack_into('<H', data, headerOffset, value)
            headerOffset += 2
        def write32(value):
            nonlocal headerOffset
            struct.pack_into('<I', data, headerOffset, value)
            headerOffset += 4

        assert headerOffset == 0
        writeRaw(self.name.ljust(12, b'\0')[:12])
        assert len(self.idCode) == 4
        writeRaw(self.idCode)
        assert len(self.developerCode) == 2
        writeRaw(self.developerCode)
        write8(self.unitCode)
        write8(self.encryptionSeedSelect)
        write8(self.deviceCapacity)
        assert headerOffset == 0x15
        write8(self.pad015)
        write8(self.pad016)
        write8(self.pad017)
        write8(self.pad018)
        write8(self.pad019)
        write8(self.pad01A)
        write8(self.pad01B)
        write8(self.pad01C)
        write8(self.region)
        write8(self.version)
        write8(self.autostart)
        assert headerOffset == 0x20
        write32(arm9Offset)
        write32(self.arm9EntryAddress)
        write32(self.arm9RamAddress)
        write32(len(self.arm9))
        write32(arm7Offset)
        write32(self.arm7EntryAddress)
        write32(self.arm7RamAddress)
        write32(len(self.arm7))
        assert headerOffset == 0x40
        write32(fntOffset)
        write32(len(fnt))
        write32(fatOffset)
        write32(len(self.files) * 8)
        write32(arm9OvTOffset)
        write32(len(self.arm9OverlayTable))
        write32(arm7OvTOffset)
        write32(len(self.arm7OverlayTable))
        assert headerOffset == 0x60
        write32(self.normalCardControlRegisterSettings)
        write32(self.secureCardControlRegisterSettings)
        write32(iconBannerOffset)
        write16(self.secureAreaChecksum)
        write16(self.secureTransferDelay)
        assert headerOffset == 0x70
        write32(self.arm9CodeSettingsPointerAddress)
        write32(self.arm7CodeSettingsPointerAddress)
        writeRaw(self.secureAreaDisable.ljust(8, b'\0')[:8])
        assert headerOffset == 0x80
        write32(rsaSignatureOffset)
        write32(0x4000)
        assert len(self.pad088) == 0x38
        writeRaw(self.pad088)
        assert len(self.nintendoLogo) == 0x9C
        writeRaw(self.nintendoLogo)
        write16(_common.crc16(self.nintendoLogo))
        write16(_common.crc16(data[0:0x15e]))
        assert headerOffset == 0x160
        write32(debugRomOffset)
        write32(len(self.debugRom))
        write32(self.debugRomAddress)
        assert len(self.pad16C) == 0x94
        writeRaw(self.pad16C)
        assert headerOffset == 0x200

        return bytes(data)


    def saveToFile(self, filePath, *, updateDeviceCapacity=False):
        """
        Generate file data representing this ROM, and save it to a
        filesystem file.
        """
        d = self.save(updateDeviceCapacity=updateDeviceCapacity)
        with open(filePath, 'wb') as f:
            f.write(d)


    def loadArm9(self):
        """
        Create a MainCodeFile object representing the main ARM9 code
        file in this ROM.
        """
        return code.MainCodeFile(self.arm9,
                                 self.arm9RamAddress,
                                 self.arm9CodeSettingsPointerAddress)


    def loadArm7(self):
        """
        Create a MainCodeFile object representing the main ARM7 code
        file in this ROM.
        """
        return code.MainCodeFile(self.arm7,
                                 self.arm7RamAddress,
                                 self.arm7CodeSettingsPointerAddress)


    def loadArm9Overlays(self, idsToLoad=None):
        """
        Create a dictionary of this ROM's ARM9 overlays.
        """
        def callback(ovID, fileID):
            return self.files[fileID]
        return code.loadOverlayTable(self.arm9OverlayTable, callback, idsToLoad)


    def loadArm7Overlays(self, idsToLoad=None):
        """
        Create a dictionary of this ROM's ARM7 overlays.
        """
        def callback(ovID, fileID):
            return self.files[fileID]
        return code.loadOverlayTable(self.arm7OverlayTable, callback, idsToLoad)


    def getFileByName(self, filename):
        """
        Return the data for the file with the given filename (path).
        This is a convenience function.
        """
        fid = self.filenames.idOf(filename)
        if fid is None:
            raise ValueError(f'Cannot find file ID of "{filename}"')
        return self.files[fid]


    def setFileByName(self, filename, data):
        """
        Replace the data for the file with the given filename (path)
        with the given data. This is a convenience function.
        """
        fid = self.filenames.idOf(filename)
        if fid is None:
            raise ValueError(f'Cannot find file ID of "{filename}"')
        self.files[fid] = data


    def __str__(self):
        title = repr(bytes(self.name))[2:-1].rstrip(' ')
        code = repr(bytes(self.idCode))[2:-1]
        return f'<rom "{title}" ({code})>'


    def __repr__(self):
        try:
            data = _common.shortBytesRepr(self.save())
        except Exception:
            data = '...'
        return f'{type(self).__name__}({data})'



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for SDAT sound archives.
"""


import itertools
import struct

from . import _common
from . import soundBank
from . import soundGroup
from . import soundSequence
from . import soundSequenceArchive
from . import soundSequencePlayer
from . import soundStream
from . import soundStreamPlayer
from . import soundWaveArchive


class SDAT:
    """
    A sound data archive file (SDAT).
    """
    def __init__(self, data=None):
        self.sequences = []
        self.sequenceArchives = []
        self.banks = []
        self.waveArchives = []
        self.sequencePlayers = []
        self.groups = []
        self.streamPlayers = []
        self.streams = []

        self.fatLengthsIncludePadding = False
        self.firstFileAlignment = None
        self.fileAlignment = 0x20
        self.padAtEnd = True # only inaccurate 3rd party tools change
                             # this from the default, but we support it
                             # anyway
        self.padSymbSizeTo4InSDATHeader = False # same as above

        if data is not None:
            self._initFromData(data)


    def _initFromData(self, data):
        """
        Initialize the SDAT using the raw file data given.
        """
        # Read the standard header
        magic, bom, version, filesize, headersize, numblocks = \
            _common.NDS_STD_FILE_HEADER.unpack_from(data, 0)
        if version != 0x100:
            raise ValueError(f'Unsupported SDAT version: {version}')

        if magic != b'SDAT':
            raise ValueError("Wrong magic (should be b'SDAT', instead found "
                             f'{magic})')

        # Read the block offsets and sizes
        (symbolsBlockOffset, symbolsBlockSize,
            infoBlockOffset, infoBlockSize,
            fatBlockOffset, fatBlockSize,
            fileBlockOffset, fileBlockSize,
            ) = struct.unpack_from('<8I', data, 0x10)

        # Read the symbols block
        (symbolsMagic, symbolsSize) = \
            struct.unpack_from('<4sI', data, symbolsBlockOffset)

        if symbolsBlockOffset != 0:
            symbolsOffsets = struct.unpack_from('<8I', data,
                                                symbolsBlockOffset + 8)
            assert symbolsMagic == b'SYMB'
        else:
            symbolsOffsets = [None] * 8


        lastEndOfString = 0 # relative to SYMB block
        def readSymbolsList(offset, hasSubgroups):
            """
            Read a list of symbols at offset offset. If hasSubgroups,
            it'll be parsed assuming that the symbol table has entries
            for sub-symbol-lists as well. (In practice, this only occurs
            for SSARs.)
            If there are no symbols, return an empty list.
            """
            nonlocal lastEndOfString

            if offset is None: return []

            off = symbolsBlockOffset + offset
            count, = struct.unpack_from('<I', data, off); off += 4

            symbols = []
            for i in range(count):
                symbolOff, = struct.unpack_from('<I', data, off)
                off += 4

                if symbolOff == 0:
                    thisSymbol = None
                else:
                    thisSymbol = _common.loadNullTerminatedStringFrom(data,
                        symbolsBlockOffset + symbolOff)
                    lastEndOfString = symbolOff + len(thisSymbol) + 1

                if not hasSubgroups:
                    symbols.append(thisSymbol)
                else:
                    subSymbolsOff, = struct.unpack_from('<I', data, off)
                    off += 4

                    if subSymbolsOff == 0:
                        subSymbols = []
                    else:
                        subSymbols = readSymbolsList(subSymbolsOff, False)

                    symbols.append((thisSymbol, subSymbols))
            
            return symbols


        # Read the FAT block
        (fatMagic, fatSize, fatCount) = \
            struct.unpack_from('<4sII', data, fatBlockOffset)
        assert fatMagic == b'FAT ' # note trailing space

        # Read the files from the FILES block
        files = []
        fatArrayPos = fatBlockOffset + 0x0C
        self.fileAlignment = 0x200
        self.fatLengthsIncludePadding = True
        finalFileEnd = fileBlockOffset + 8
        for i in range(fatCount):
            (fileOffset, fileSize) = \
                struct.unpack_from('<II', data, fatArrayPos)
            fatArrayPos += 0x10 # There's 8 pad bytes.

            # We'll need this later
            finalFileEnd = fileOffset + fileSize

            if i != fatCount - 1:
                nextOffset, = struct.unpack_from('<I', data, fatArrayPos)
                paddedSize = nextOffset - fileOffset
                if paddedSize != fileSize:
                    self.fatLengthsIncludePadding = False

            # Most SDATs require files to be padded to 0x20, but some
            # use other amounts. We check for that here, so that we can
            # rebuild it correctly later.
            if fileOffset % 0x200 == 0x100:
                self.fileAlignment = min(self.fileAlignment, 0x100)
            if fileOffset % 0x100 == 0x80:
                self.fileAlignment = min(self.fileAlignment, 0x80)
            if fileOffset % 0x80 == 0x40:
                self.fileAlignment = min(self.fileAlignment, 0x40)
            if fileOffset % 0x40 == 0x20:
                self.fileAlignment = min(self.fileAlignment, 0x20)
            if fileOffset % 0x20 == 0x10:
                self.fileAlignment = min(self.fileAlignment, 0x10)
            if fileOffset % 0x10 == 8:
                self.fileAlignment = min(self.fileAlignment, 8)
            if fileOffset % 8 == 4:
                self.fileAlignment = min(self.fileAlignment, 4)
            if fileOffset % 4 == 2:
                self.fileAlignment = min(self.fileAlignment, 2)
            if fileOffset % 2 == 1: # yes, this happens sometimes
                self.fileAlignment = min(self.fileAlignment, 1)

            if i == 0:
                self.firstFileAlignment = self.fileAlignment

            file = data[fileOffset : fileOffset + fileSize]
            files.append(file)

        if self.firstFileAlignment == self.fileAlignment:
            self.firstFileAlignment = None

        # Check if the end is definitely unpadded (that is, if there
        # should be padding and it's not present)
        if finalFileEnd == len(data) and finalFileEnd % self.fileAlignment != 0:
            self.padAtEnd = False

        # Do another quick pass to find if the FAT file lengths include
        # padding

        # Read the info block
        (infoMagic, infoSize) = \
            struct.unpack_from('<4sI', data, infoBlockOffset)
        infoOffsets = struct.unpack_from('<8I', data,
                                         infoBlockOffset + 8)
        assert infoMagic == b'INFO'

        def getInfoEntryOffsets(partNum):
            off = infoOffsets[partNum]
            count, = struct.unpack_from('<I', data, infoBlockOffset + off)
            entryOffsets = struct.unpack_from(f'<{count}I', data,
                                              infoBlockOffset + off + 4)
            for entryOff in entryOffsets:
                if entryOff == 0:
                    yield None
                else:
                    yield infoBlockOffset + entryOff


        # Info part 0: SSEQ (references SBNK)
        for entryOff, symb in itertools.zip_longest(getInfoEntryOffsets(0),
                                  readSymbolsList(symbolsOffsets[0], False)):
            if entryOff is None:
                sseq = None
            else:
                (fileID, unk02, bankID, volume, channelPressure,
                    polyphonicPressure, playerID) = \
                    struct.unpack_from('<3H4B', data, entryOff)
                sseq = soundSequence.SSEQ(files[fileID], unk02, bankID,
                            volume, channelPressure, polyphonicPressure, playerID)
                sseq.dataMergeOptimizationID = fileID

            self.sequences.append((symb, sseq))

        # Info part 1: SSAR
        for entryOff, symb in itertools.zip_longest(getInfoEntryOffsets(1),
                                  readSymbolsList(symbolsOffsets[1], True)):
            if entryOff is None:
                ssar = None
            else:
                fileID, unk02 = struct.unpack_from('<HH', data, entryOff)
                subSymb = symb[1] if symb is not None else None
                ssar = soundSequenceArchive.SSAR(files[fileID], unk02, subSymb)
                ssar.dataMergeOptimizationID = fileID

            name = symb[0] if symb is not None else None
            self.sequenceArchives.append((name, ssar))

        # Info part 2: SBNK
        for entryOff, symb in itertools.zip_longest(getInfoEntryOffsets(2),
                                  readSymbolsList(symbolsOffsets[2], False)):
            if entryOff is None:
                sbnk = None
            else:
                fileID, unk02 = struct.unpack_from('<HH', data, entryOff)
                swarIDs = struct.unpack_from('<4h', data, entryOff + 4)
                swarIDs2 = []
                for x in swarIDs:
                    if x == -1:
                        swarIDs2.append(None)
                    else:
                        swarIDs2.append(x)

                sbnk = soundBank.SBNK(files[fileID], unk02, swarIDs2)
                sbnk.dataMergeOptimizationID = fileID

            self.banks.append((symb, sbnk))

        # Info part 3: SWAR
        for entryOff, symb in itertools.zip_longest(getInfoEntryOffsets(3),
                                  readSymbolsList(symbolsOffsets[3], False)):
            if entryOff is None:
                swar = None
            else:
                fileID, unk02 = struct.unpack_from('<HH', data, entryOff)
                swar = soundWaveArchive.SWAR(files[fileID], unk02)
                swar.dataMergeOptimizationID = fileID

            self.waveArchives.append((symb, swar))

        # Info part 4: Sequence players
        for entryOff, symb in itertools.zip_longest(getInfoEntryOffsets(4),
                                  readSymbolsList(symbolsOffsets[4], False)):
            if entryOff is None:
                sp = None
            else:
                maxSequences, channelMask, heapSize = \
                    struct.unpack_from('<HHI', data, entryOff)

                channels = set()
                for i in range(16):
                    if (channelMask >> i) & 1:
                        channels.add(i)

                sp = soundSequencePlayer.SequencePlayer(maxSequences,
                                                   channels,
                                                   heapSize)

            self.sequencePlayers.append((symb, sp))

        # Info part 5: Groups
        for groupOff, symb in itertools.zip_longest(getInfoEntryOffsets(5),
                                  readSymbolsList(symbolsOffsets[5], False)):
            if groupOff is None:
                entries = None
            else:
                entriesCount, = struct.unpack_from('<I', data, groupOff)

                entries = []
                arrayOff = groupOff + 4
                for i in range(entriesCount):
                    type, options, id = struct.unpack_from('<BHxI', data, arrayOff)
                    arrayOff += 8

                    entries.append(soundGroup.GroupEntry(type, options, id))

            self.groups.append((symb, entries))

        # Info part 6: Stream players
        for entryOff, symb in itertools.zip_longest(getInfoEntryOffsets(6),
                                  readSymbolsList(symbolsOffsets[6], False)):
            if entryOff is None:
                sp = None
            else:
                count, = struct.unpack_from('<B', data, entryOff)
                channels = list(
                    struct.unpack_from(f'<{count}B', data, entryOff + 1))
                sp = soundStreamPlayer.StreamPlayer(channels)

            self.streamPlayers.append((symb, sp))

        # Info part 7: Streams
        for entryOff, symb in itertools.zip_longest(getInfoEntryOffsets(7),
                                  readSymbolsList(symbolsOffsets[7], False)):
            if entryOff is None:
                strm = None
            else:
                fileID, unk02, volume, priority, playerID, unk07 = \
                    struct.unpack_from('<HH4B', data, entryOff)
                strm = soundStream.STRM(files[fileID], unk02, volume, priority, playerID, unk07)
                strm.dataMergeOptimizationID = fileID

            self.streams.append((symb, strm))


        # If the symbols block size is definitely padded, record that
        if symbolsBlockSize % 4 == 0 and lastEndOfString % 4 != 0:
            self.padSymbSizeTo4InSDATHeader = True


    @classmethod
    def fromFile(cls, filePath):
        """
        Load an SDAT from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read())


    def save(self):
        """
        Generate file data representing this SDAT.
        """
        # First, just allocate enough memory for the SDAT header.
        data = bytearray(0x40)

        # -------------------
        # Make the SYMB block

        symbolsStringTable = bytearray()
        def addSymbolAndGetOffset(symbol):
            if symbol is None:
                return -1
            offset = len(symbolsStringTable)
            symbolsStringTable.extend(symbol.encode('latin-1') + b'\0')
            return offset

        symbolsHeaderOffsets = []

        # Parallel arrays, here.
        symbolsTableValues = []
        shouldIncrementByTableLen = []

        anySymbolsInWholeFile = False

        def addSymbolsFrom(namedList, nested=False):

            # First, figure out if any actual symbols exist
            anyActualSymbols = False
            anyActualSubsymbols = False
            if not nested:
                for symbol, _ in namedList:
                    if symbol is not None:
                        anyActualSymbols = True
                        break
            else:
                for symbol, entry in namedList:
                    if symbol is not None:
                        anyActualSymbols = True
                        break
                    for subSymbol, subEntry in entry.sequences:
                        if subSymbol is not None:
                            anyActualSubsymbols = True
                            break


            nonlocal anySymbolsInWholeFile
            anySymbolsInWholeFile |= anyActualSymbols
            anySymbolsInWholeFile |= anyActualSubsymbols

            # If there *are* any symbols, keep going
            symbolsHeaderOffsets.append(len(symbolsTableValues) * 4)

            if not nested:
                symbolsTableValues.append(len(namedList))
                shouldIncrementByTableLen.append(False)

                for symbol, _ in namedList:
                    symbolsTableValues.append(addSymbolAndGetOffset(symbol))
                    shouldIncrementByTableLen.append(True)

            else:
                mainList, subListsArea = [], []
                mainListSIBTL, subListsAreaSIBTL = [], []

                mainList.append(len(namedList))
                mainListSIBTL.append(False)

                mainListFullLength = (1 + 2 * len(namedList)) * 4
                subListsAreaOffset = (0x40
                                      + len(symbolsTableValues) * 4
                                      + mainListFullLength)

                for symbol, entry in namedList:

                    mainList.append(addSymbolAndGetOffset(symbol))
                    mainListSIBTL.append(True)

                    subListOffset = subListsAreaOffset + len(subListsArea) * 4

                    if entry is None:
                        subNames = []
                    else:
                        subNames = [n for (n, s) in entry.sequences]

                    if entry or subNames:
                        subListsArea.append(len(subNames))
                        subListsAreaSIBTL.append(False)

                        for subSymbol in subNames:
                            subListsArea.append(addSymbolAndGetOffset(subSymbol))
                            subListsAreaSIBTL.append(True)

                        mainList.append(subListOffset)
                        mainListSIBTL.append(False)

                    else:
                        mainList.append(0)
                        mainListSIBTL.append(False)

                symbolsTableValues.extend(mainList)
                symbolsTableValues.extend(subListsArea)
                shouldIncrementByTableLen.extend(mainListSIBTL)
                shouldIncrementByTableLen.extend(subListsAreaSIBTL)

        addSymbolsFrom(self.sequences)
        addSymbolsFrom(self.sequenceArchives, True)
        addSymbolsFrom(self.banks)
        addSymbolsFrom(self.waveArchives)
        addSymbolsFrom(self.sequencePlayers)
        addSymbolsFrom(self.groups)
        addSymbolsFrom(self.streamPlayers)
        addSymbolsFrom(self.streams)

        # Only add the SYMB block if there are any symbols
        if anySymbolsInWholeFile:
            symbolsBlockOffset = len(data)

            symbolsTableLen = len(symbolsTableValues) * 4
            symbolsTable = bytearray()
            for value, shouldIncrement in itertools.zip_longest(symbolsTableValues,
                                              shouldIncrementByTableLen):
                if value == -1:
                    symbolsTable.extend(b'\0\0\0\0')
                else:
                    if shouldIncrement:
                        value += symbolsTableLen + 0x40
                    symbolsTable.extend(struct.pack('<I', value))

            symbolsBlockSize = 0x40 + len(symbolsTable) + len(symbolsStringTable)
            paddedSymbSize = symbolsBlockSize
            while paddedSymbSize % 4:
                paddedSymbSize += 1
            if self.padSymbSizeTo4InSDATHeader:
                symbolsBlockSize = paddedSymbSize

            symbolsHeaderOffsetsTable = bytearray()
            for value in symbolsHeaderOffsets:
                if value is None:
                    symbolsHeaderOffsetsTable.extend(b'\0\0\0\0')
                else:
                    symbolsHeaderOffsetsTable.extend(struct.pack('<I', value + 0x40))

            symbolsHeader = struct.pack('<4sI',
                b'SYMB', paddedSymbSize)

            data.extend(symbolsHeader)
            data.extend(symbolsHeaderOffsetsTable)
            data.extend(b'\0' * 0x18)
            data.extend(symbolsTable)
            data.extend(symbolsStringTable)

        else:
            symbolsBlockOffset = None
            symbolsBlockSize = None


        # -------------------
        # Make the INFO block
        while len(data) % 4: data.append(0)
        infoBlockOffset = len(data)

        # Add room to add the header later
        data.extend(b'\0' * (8 + 8 * 4))

        # Pad to 0x20 relative to the INFO block, for some reason
        while (len(data) - infoBlockOffset) % 0x20: data.append(0)

        # Helper functions
        def info_declarePart(partNumber):
            struct.pack_into('<I', data, infoBlockOffset + 8 + 4 * partNumber,
                len(data) - infoBlockOffset)
        def addFileAndGetID(file, dataMergeOptimizationID):
            idx = _common.listFind(files, file)

            while idx != -1:
                if dataMergeOptimizationID == fileMergeIDs[idx]:
                    return idx
                idx = _common.listFind(files, file, idx + 1)

            files.append(file)
            fileMergeIDs.append(dataMergeOptimizationID)
            return len(files) - 1

        # We encode sections out of order, so that the files will be in
        # the same order as in retail SDATs.
        fileMergeIDs = []
        files = []

        # Info part 0: SSEQ
        info_declarePart(0)

        data.extend(struct.pack('<I', len(self.sequences)))
        sseqOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.sequences)))

        for i, (_, sseq) in enumerate(self.sequences):
            if sseq is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                (file, unk02, bankID, volume, channelPressure,
                    polyphonicPressure, playerID) = sseq.save()
                fileID = addFileAndGetID(file, sseq.dataMergeOptimizationID)

                data.extend(struct.pack('<3H4Bxx',
                    fileID, unk02, bankID, volume, channelPressure,
                    polyphonicPressure, playerID))

            struct.pack_into('<I', data, sseqOffsetsTableOffset + 4 * i, entryOff)

        # Info part 1: SSAR
        info_declarePart(1)

        data.extend(struct.pack('<I', len(self.sequenceArchives)))
        ssarOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.sequenceArchives)))

        for i, (_, ssar) in enumerate(self.sequenceArchives):
            if ssar is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                file, unk02, _ = ssar.save()
                fileID = addFileAndGetID(file, ssar.dataMergeOptimizationID)

                data.extend(struct.pack('<HH',
                    fileID, unk02))

            struct.pack_into('<I', data, ssarOffsetsTableOffset + 4 * i, entryOff)

        # Info part 2: SBNK
        info_declarePart(2)

        data.extend(struct.pack('<I', len(self.banks)))
        sbnkOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.banks)))

        for i, (sbnkName, sbnk) in enumerate(self.banks):
            if sbnk is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                file, unk02, waveArchives = sbnk.save()
                fileID = addFileAndGetID(file, sbnk.dataMergeOptimizationID)

                swarIDs = []
                for s in waveArchives:
                    swarIDs.append(-1 if s is None else s)
                while len(swarIDs) < 4:
                    swarIDs.append(-1)

                if len(swarIDs) > 4:
                    raise ValueError(f'SBNK {i} ("{sbnkName}") uses '
                        f'{len(swarIDs)} SWARs. The maximum is 4.')

                data.extend(struct.pack('<HH4h',
                    fileID, unk02, *swarIDs))

            struct.pack_into('<I', data, sbnkOffsetsTableOffset + 4 * i, entryOff)


        # Info part 3: SWAR
        info_declarePart(3)

        data.extend(struct.pack('<I', len(self.waveArchives)))
        swarOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.waveArchives)))

        for i, (_, swar) in enumerate(self.waveArchives):
            if swar is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                file, unk02 = swar.save()
                fileID = addFileAndGetID(file, swar.dataMergeOptimizationID)

                data.extend(struct.pack('<HH',
                    fileID, unk02))

            struct.pack_into('<I', data, swarOffsetsTableOffset + 4 * i, entryOff)


        # Info part 4: Sequence players
        info_declarePart(4)

        data.extend(struct.pack('<I', len(self.sequencePlayers)))
        spOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.sequencePlayers)))

        for i, (_, sp) in enumerate(self.sequencePlayers):
            if sp is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                maxSequences, channels, heapSize = sp.save()

                channelMask = 0
                for j in range(16):
                    if j in channels:
                        channelMask |= 1 << j

                data.extend(struct.pack('<HHI',
                    maxSequences, channelMask, heapSize))

            struct.pack_into('<I', data, spOffsetsTableOffset + 4 * i, entryOff)


        # Info part 5: Groups
        info_declarePart(5)

        data.extend(struct.pack('<I', len(self.groups)))
        groupOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.groups)))

        for i, (_, group) in enumerate(self.groups):
            if group is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                data.extend(struct.pack('<I', len(group)))

                for gEntry in group:
                    data.extend(struct.pack('<BHxI', *gEntry.save()))

            struct.pack_into('<I', data, groupOffsetsTableOffset + 4 * i, entryOff)


        # Info part 6: Stream players
        info_declarePart(6)

        data.extend(struct.pack('<I', len(self.streamPlayers)))
        spOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.streamPlayers)))

        for i, (_, sp) in enumerate(self.streamPlayers):
            if sp is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                channels = sp.save()
                chanCount = len(channels)
                while len(channels) < 16:
                    channels.append(0xFF)

                data.append(chanCount)
                data.extend(channels)

                # This has to occur in order for the padding to work out
                # correctly. Weird, but, what can you do. Might even be
                # an unknown value.
                data.extend(b'\0\0\0\0')

            struct.pack_into('<I', data, spOffsetsTableOffset + 4 * i, entryOff)

            while len(data) % 4: data.append(0)


        # Info part 7: Streams
        info_declarePart(7)

        data.extend(struct.pack('<I', len(self.streams)))
        strmOffsetsTableOffset = len(data)
        data.extend(b'\0' * (4 * len(self.streams)))

        for i, (_, strm) in enumerate(self.streams):
            if strm is None:
                entryOff = 0
            else:
                entryOff = len(data) - infoBlockOffset

                file, unk02, volume, priority, playerID, unk07 = strm.save()
                fileID = addFileAndGetID(file, strm.dataMergeOptimizationID)

                data.extend(struct.pack('<HH4B4x',
                    fileID, unk02, volume, priority, playerID, unk07))

            struct.pack_into('<I', data, strmOffsetsTableOffset + 4 * i, entryOff)

        # Now we can finally fill the header in.
        struct.pack_into('<4sI', data, infoBlockOffset,
            b'INFO', len(data) - infoBlockOffset)

        infoBlockSize = len(data) - infoBlockOffset


        # ----------------------
        # Make a dummy FAT block, to be filled in when adding to the
        # FILE block

        while len(data) % 4: data.append(0)
        fatBlockOffset = len(data)
        fatBlockSize = 0xC + 0x10 * len(files)
        fatTableOffset = fatBlockOffset + 0xC

        fatHeader = struct.pack('<4sII',
            b'FAT ', 0xC + 0x10 * len(files), len(files))

        data.extend(fatHeader)
        data.extend(b'\0' * (0x10 * len(files)))


        # -------------------
        # Make the FILE block and fill in the FAT block
        while len(data) % 4: data.append(0)
        fileBlockOffset = len(data)

        # Dummy header (to be filled in after we know the total size)
        data.extend(b'\0' * 0xC)

        # Some games align the first file differently
        if self.firstFileAlignment is not None:
            while len(data) % self.firstFileAlignment:
                data.append(0)

        # Add each file
        for i, file in enumerate(files):

            # Files must be aligned to 0x20 relative to the SDAT
            # itself... usually. Some games align to other amounts.
            while len(data) % self.fileAlignment:
                data.append(0)

            # Actually add the file
            fileOffset = len(data)
            data.extend(file)
            
            # Add the appropriate FAT entry
            fLen = len(file)
            if self.fatLengthsIncludePadding:
                while fLen % self.fileAlignment: fLen += 1

            struct.pack_into('<II', data, fatTableOffset + 0x10 * i,
                fileOffset, fLen)

        # And one last pad for good measure. (And because retail files
        # do so.)
        if self.padAtEnd:
            while len(data) % self.fileAlignment:
                data.append(0)

        # Add the header
        struct.pack_into('<4sII', data, fileBlockOffset,
            b'FILE', len(data) - fileBlockOffset, len(files))

        fileBlockSize = len(data) - fileBlockOffset


        # -----------------------
        # Put the blocks together

        # Write the SDAT header
        struct.pack_into('<8I', data, 0x10,
            0 if symbolsBlockOffset is None else symbolsBlockOffset,
            0 if symbolsBlockSize is None else symbolsBlockSize,
            0 if infoBlockOffset is None else infoBlockOffset,
            0 if infoBlockSize is None else infoBlockSize,
            0 if fatBlockOffset is None else fatBlockOffset,
            0 if fatBlockSize is None else fatBlockSize,
            0 if fileBlockOffset is None else fileBlockOffset,
            0 if fileBlockSize is None else fileBlockSize)

        # Write the standard header to the beginning
        _common.NDS_STD_FILE_HEADER.pack_into(data, 0,
            b'SDAT', 0xFEFF, 0x100, len(data), 0x40,
            3 if symbolsBlockOffset is None else 4)

        return data


    def saveToFile(self, filePath):
        """
        Generate file data representing this SDAT, and save it to a
        filesystem file.
        """
        d = self.save()
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        return (f'<sdat'
                f' {len(self.sequences)} sequences'
                f' {len(self.sequenceArchives)} sequence archives'
                f' {len(self.banks)} banks'
                f' {len(self.waveArchives)} wave archives'
                f' {len(self.sequencePlayers)} sequence players'
                f' {len(self.groups)} groups'
                f' {len(self.streamPlayers)} stream players'
                f' {len(self.streams)} streams>')


    def __repr__(self):
        try:
            data = _common.shortBytesRepr(self.save())
        except Exception:
            data = '...'
        return f'{type(self).__name__}({data})'



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for executable code files compression.
"""


import struct

from . import _lzCommon


def _detectAppendedData(data):
    """
    Attempt to check if there's any appended data at the end of the
    given data. Returns an integer representing the amount of such data
    if so, or None if the data doesn't seem to be compressed.
    """
    for possibleAmt in range(0, 0x20, 4):
        try:
            headerLen = data[-5 - possibleAmt]
        except IndexError:
            return None

        compLenHeaderLen, extraSize = struct.unpack_from('<II', data, len(data) - possibleAmt - 8)
        headerLen = compLenHeaderLen >> 24
        compressedLen = compLenHeaderLen & 0xFFFFFF

        if headerLen < 8: continue
        if compressedLen > len(data): continue
        for byte in data[-possibleAmt - headerLen  : -possibleAmt - 8]:
            if byte != 0xFF:
                continue

        return possibleAmt


def decompress(data):
    """
    Decompress data that was compressed using code compression. This is
    the inverse of compress().
    """

    # NOTE:
    # This code is ported from DSDecmp:
    # http://www.romhacking.net/utilities/789/
    # https://github.com/Barubary/dsdecmp

    appendedDataAmount = _detectAppendedData(data)

    if appendedDataAmount is None:
        # Probably not compressed.
        return data

    if appendedDataAmount == 0:
        appendedData = b''
    else:
        appendedData = data[-appendedDataAmount:]
        data = data[:-appendedDataAmount]

    # If extraSize (the last value in the header) is 0, the data is not
    # actually compressed
    if data[-4:] == b'\0\0\0\0':
        return data

    # Read the header
    compLenHeaderLen, extraSize = struct.unpack_from('<II', data, len(data) - 8)
    headerLen = compLenHeaderLen >> 24
    compressedLen = compLenHeaderLen & 0xFFFFFF

    if len(data) < headerLen:
        raise ValueError(f'File is too small for header ({len(data)} < {headerLen})')

    if compressedLen > len(data):
        raise ValueError("Compressed length doesn't fit in the input file")

    # The rest of the header should be all 0xFF's
    for byte in data[-headerLen:-8]:
        if byte != 0xFF:
            raise ValueError("Header padding isn't entirely 0xFF")


    # Format description:
    #
    # Code LZ compression is basically just LZ-0x10 compression.
    # However, the order of reading is reversed: the compression starts
    # at the end of the file. Assuming we start reading at the end
    # towards the beginning, the format is:
    #
    # u32 extraSize            | decompressed data size = file length
    #                          | (including header) + this value
    # u8 headerLen             |
    # u24 compressedLen        | Can be less than file size (without
    #                          | header): if so, the rest of the file is
    #                          | uncompressed. It may also be the file
    #                          | size.
    # u8[headerSize-8] padding | 0xFF's
    # 
    # 0x10-like-compressed data follows (without the usual 4-byte
    # header). The only difference is that 2 should be added to the DISP
    # value in compressed blocks to get the proper value. The u32 and
    # u24 are read most significant byte first. If extraSize is 0, there
    # is no headerSize, decompressedLength or padding: the data starts
    # immediately, and is uncompressed.
    # 
    # arm9.bin has 3 extra u32 values at the 'start' (ie: end of the
    # file), which may be ignored (and are ignored here). These 12 bytes
    # also should not be included in the computation of the output size.

    # The compressed size is sometimes the file size
    if compressedLen >= len(data):
        compressedLen = len(data)

    # The first part of the file, not included in compressedLen, is not
    # compressed, and should be ignored.
    passthroughLen = len(data) - compressedLen
    passthroughData = data[:passthroughLen]

    # Then there's the compressed data. Also make a bytearray where
    # we'll be putting the decompressed data.
    compData = data[passthroughLen : passthroughLen + compressedLen - headerLen]
    decompData = bytearray(len(data) + extraSize - passthroughLen)

    # Main decompression loop
    currentOutSize = 0
    decompLen = len(decompData)
    readBytes = 0
    flags = 0
    mask = 1
    while currentOutSize < decompLen:

        # Update the mask. If all flag bits have been read, get a new
        # set.
        if mask == 1:
            if readBytes >= compressedLen:
                raise RuntimeError('Not enough data to decompress')

            flags = compData[-1 - readBytes]
            readBytes += 1

            mask = 0x80
        else:
            mask >>= 1

        # Bit = 1 means it's compressed
        if flags & mask:
            # Get length and displacement ("disp") values from the next 2 bytes
            if readBytes + 1 >= len(data):
                raise RuntimeError('Not enough data to decompress')

            byte1 = compData[-1 - readBytes]; readBytes += 1
            byte2 = compData[-1 - readBytes]; readBytes += 1

            # The number of bytes to copy
            length = (byte1 >> 4) + 3

            # Where the bytes should be copied from (relatively)
            disp = (((byte1 & 0x0F) << 8) | byte2) + 3

            if disp > currentOutSize:
                if currentOutSize < 2:
                    raise RuntimeError(
                        'Cannot go back more than already written; '
                        f'attempted to go back {hex(disp)} bytes when only '
                        f'{hex(currentOutSize)} bytes have been written')

                # HACK. This seems to produce valid files, but isn't the
                # most elegant solution. Although this *could* be the
                # actual way to use a disp of 2 in this format, as,
                # otherwise, the minimum would be 3 (and 0 is undefined,
                # and 1 is less useful).
                disp = 2

            bufIdx = currentOutSize - disp
            for i in range(length):
                next = decompData[-1 - bufIdx]
                bufIdx += 1

                decompData[-1 - currentOutSize] = next
                currentOutSize += 1

        else:
            if readBytes > len(data):
                raise RuntimeError('Not enough data to decompress')

            next = compData[-1 - readBytes]
            readBytes += 1

            decompData[-1 - currentOutSize] = next
            currentOutSize += 1

    return passthroughData + decompData + appendedData


def compress(data, isArm9=False):
    """
    Compress code data. This is the inverse of decompress().
    """

    # NOTE:
    # This code is ported (with substantial changes) from
    # Nintendo DS/GBA Compressors by CUE:
    # https://gbatemp.net/threads/nintendo-ds-gba-compressors.313278/
    # http://www.romhacking.net/utilities/826/

    if isArm9:
        prefix = data[:0x4000]
        data = data[0x4000:]
    else:
        prefix = b''

    return bytes(prefix + _compress(data))


def _compress(data):

    compressed, ignorableD, ignorableC = \
        _lzCommon.compress(bytes(reversed(data)), 3, 0x1002, 18, False, True)
    compressed = bytearray(reversed(compressed))

    if not compressed or (len(data) + 4 < ((len(compressed) + 3) & ~4) + 8):
        # Compressed size too large -- copy the raw data over instead

        compressed = bytearray(data)

        while len(compressed) % 4:
            compressed.append(0)

        for i in range(4):
            compressed.append(0)

        return compressed

    else:
        # Remove ignorable data (i.e. data that can remain uncompressed
        # without making the overall output file size larger), and tack
        # on the header at the end

        actualCompLen = len(compressed) - ignorableC
        headerLen = 8
        compressed = data[:ignorableD] + compressed[ignorableC:]
        extraLen = len(data) - len(compressed)

        while len(compressed) % 4:
            compressed.append(0xFF)
            headerLen += 1

        ptr = len(compressed)

        compressed.extend(b'\0' * 8)
        struct.pack_into('<I', compressed, ptr, actualCompLen + headerLen)
        compressed[ptr + 3] = headerLen
        ptr += 4
        struct.pack_into('<I', compressed, ptr, extraLen - headerLen)

    return compressed
